#ifndef _VIADESC_H
#define _VIADESC_H
/*++

	Tandem Confidential: Need to Know Only

	Copyright (C) 1997, Tandem Computers Incorporated
	Protected as an unpublished work.
	All rights reserved.

	The computer program listings, specifications and documentation
	herein are the property of Tandem Computers Incorporated or a third
	party supplier and shall not be reproduced, copied, disclosed, or
	used in whole or in part for any reason without the prior express
	written permission of Tandem Computers Incorporated.

Project Name:

    VIA Emulator

File Name:

    viadesc.h

Abstract:

    This file contains structure definitions defined in the VIA
    Spec that are hardware aware. The clients, the User Agent, the Kernel 
    Agent and the Hardware all need access to these structures.

Author:

	VIA Emulator Team
    
Revision History:

	12/18/1997	SC 		split from vipl.h 
	01/28/1998	JRH		Corrected misspelling of VIP_CONTROL_IMMEDIATE.
	03/27/1998	JRH		Added convenient masks for determining whether
						or not any of the reserved bits in the control
						field of a control segment is on, and also for
						masking off the status field of a control
						segment to see if there are any errors.
	04/03/1998	JRH		Converted to new descriptor format.
	05/13/1998	JRH		Added new definition for the necessary descriptor
						starting address alignment.

--*/


#ifdef __cplusplus
extern "C" {
#endif

#include "viatype.h"

/*
 * The address alignment boundary constraint for VIA descriptors.
 */
#define VIP_DESCRIPTOR_ALIGNMENT	64

/**********************************************************************
 * Data Types
 **********************************************************************/
/* 
 * Control segment of a VIA descriptor 
 */
typedef struct {
	VIP_PVOID64		Next;
	VIP_MEM_HANDLE	NextHandle;
	VIP_UINT16		SegCount;
	VIP_UINT16		Control;
	VIP_UINT32		Reserved;
	VIP_UINT32		ImmediateData;
	VIP_UINT32		Length;
	VIP_UINT32		Status;
} VIP_CONTROL_SEGMENT, *PVIP_CONTROL_SEGMENT;

/*
 * Possible values for the Control field of the VIP_CONTROL_SEGMENT.
 */
#define VIP_CONTROL_OP_SENDRECV			0x0000
#define VIP_CONTROL_OP_RDMAWRITE		0x0001
#define VIP_CONTROL_OP_RDMAREAD			0x0002
#define VIP_CONTROL_IMMEDIATE			0x0004
#define VIP_CONTROL_QFENCE				0x0008
#define VIP_CONTROL_RESERVED			0xFFF0

/*
 * Values associated with the Status field of the VIP_CONTROL_SEGMENT.
 */
#define VIP_STATUS_DONE					0x00000001
#define VIP_STATUS_FORMAT_ERROR			0x00000002
#define VIP_STATUS_PROTECTION_ERROR		0x00000004
#define VIP_STATUS_LENGTH_ERROR			0x00000008
#define VIP_STATUS_PARTIAL_ERROR		0x00000010
#define VIP_STATUS_DESC_FLUSHED_ERROR	0x00000020
#define VIP_STATUS_TRANSPORT_ERROR		0x00000040
#define VIP_STATUS_RDMA_PROT_ERROR		0x00000080
#define VIP_STATUS_REMOTE_DESC_ERROR	0x00000100
#define VIP_STATUS_ERROR_MASK			0x000001FE
#define VIP_STATUS_OP_SEND				0x00000000
#define VIP_STATUS_OP_RECEIVE			0x00010000
#define VIP_STATUS_OP_RDMA_WRITE		0x00020000
#define VIP_STATUS_OP_REMOTE_RDMA_WRITE	0x00030000
#define VIP_STATUS_OP_RDMA_READ			0x00040000
#define VIP_STATUS_OP_MASK				0x00070000
#define VIP_STATUS_IMMEDIATE			0x00080000
#define VIP_STATUS_RESERVED				0xFFF0FE00

/*
 * VIP_ADDRESS_SEGMENT describes the remote address portion of an
 * RDMA read/write descriptor.
 */
typedef struct {
	VIP_PVOID64		Data;
	VIP_MEM_HANDLE	Handle;
	VIP_UINT32		Reserved;
} VIP_ADDRESS_SEGMENT, *PVIP_ADDRESS_SEGMENT;

/*
 * VIP_DATA_SEGMENT describes the data portion of a descriptor.
 */
typedef struct {
	VIP_PVOID64		Data;
	VIP_MEM_HANDLE	Handle;
	VIP_UINT32		Length;
} VIP_DATA_SEGMENT, *PVIP_DATA_SEGMENT;

/*
 * VIP_DESCRIPTOR is a fairly minimal descriptor definition. 
 * In general, the DS portion of the descriptor could be a 
 * larger array.
 */
typedef union {
	VIP_ADDRESS_SEGMENT	Remote;
	VIP_DATA_SEGMENT	Local;
} VIP_DESCRIPTOR_SEGMENT;

#ifdef __USLC__

/*
 * We need to do something disgusting here to cure a namespace
 * pollution problem.  <sys/regset.h> is included by <thread.h>, which
 * is needed by a number of modules that also include viadesc.h.
 * Unfortunately, <sys/regset.h> defines preprocessor symbols named
 * after registers, including CS and DS, which conflict with the field
 * names of VIP_DESCRIPTOR.  Since CS and DS are used extensively in
 * Intel's Developers Guide for VIA, it is unrealistic to expect
 * developers to adapt to different names.  Consequently we include
 * <sys/regset.h> here whether we need it or not (with the side effect
 * of defining <sys/regset.h>'s multiple inclusion guard), then
 * undefine CS and DS so that the VIP_DESCRIPTOR declaration will
 * compile.  In this way we ensure that the names CS and DS have the
 * meaning we require whether the application code includes <thread.h>
 * before or after viadesc.h.  Since VIA code is by definition new
 * code which should use the names R_CS and R_DS instead of CS and DS
 * if they need to access these names at all (see the comments in
 * <sys/regset.h>), undefining these symbols here should be harmless.
 */

#include <sys/regset.h>
#undef CS			/* collision between VIP_DESCRIPTOR and CS in regset.h */
#undef DS			/* collision between VIP_DESCRIPTOR and DS in regset.h */

#endif /* __USLC__ */

typedef struct _VIP_DESCRIPTOR {
	VIP_CONTROL_SEGMENT		CS;
	VIP_DESCRIPTOR_SEGMENT	DS[2];
} VIP_DESCRIPTOR, *PVIP_DESCRIPTOR;


#ifdef __cplusplus
} /* End of C wrapper */
#endif

#endif	/* _VIADESC_H */

