package EvCmdMgr;

use strict;
use Exporter;
use ArrayTools;
use POSIX ":sys_wait_h";
use StandardTools;
use eNotifyGuts;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT = qw(&executeCmd &getCmdStatus &setMaxCmds &executePendingCmds);

use vars qw($maxCmds $numCmds %currentCmds @pendingCmds @deadCmds);

BEGIN
{
    $SIG{CHLD} =  \&handleChildExit;

}

my $maxCmds = 25;
my $numCmds = 0;
my %currentCmds;           # pid=>cmdID
my @pendingCmds;                # associative array of cmdID=>cmdStr pairs
my @deadCmds;                   # associative array of pid=>exitStatus
my @cmdStatus = ();             # associative array of cmdID=>exitStatus
#
# handleChildExit()
#
# signal handler for child termination processing.
#
sub handleChildExit 
{
    my $pid = 0;

    while( ($pid = waitpid(-1, &WNOHANG)) >0 ) {
        # extract the exit status,
        my $exitCode = $? >> 8;
        # and save it for later.
#        print "HCE: pid= $pid, status= $exitCode\n";
        push (@deadCmds, $pid, $exitCode);
        --$numCmds;
    }

    $SIG{CHLD} =  \&handleChildExit; # reinstate handler
}


#
# executeCmd(cmdID, cmdStr)
#
# External interface for requesting cmd execution.
#
sub executeCmd ($$$$)
{
    # cmdID, cmdStr
    my ($cmdID, $cmdStr, $notifyFlag, $syncNotifyFlag) = @_;

    # Execute previously pending cmds if any.
    executePendingCmds();

    # Do synchronous notify now, if specified as such.
    if ($notifyFlag && $syncNotifyFlag) {
        doSynchNotify($cmdID, $cmdStr);
        return;
    }

    # Proceed with asynchronous execution

    # Verify if cmd can be executed at this time.
    if ($numCmds >= $maxCmds) { # Too many outstanding cmds at this time?
        # Save for later
        push (@pendingCmds,  $cmdID, $cmdStr);
    } else {
        # Initiate cmd execution
        doCmdExecution($cmdID, $cmdStr);
    }
}

#
# doCmdExecution(cmdID, cmdStr)
#
# Handle asynchronous cmd execution.
#
sub doCmdExecution ($$)
{
    my ($cmdID, $cmdStr) = @_;
    my $pid;

    if ($debug) {
        print ERRLOG "EvCmdMgr::doCmdExecution=> Executing cmd: cmdID= $cmdID, cmdStr= $cmdStr\n";
    }

    if ($pid = fork()) {
        # parent
        # Save child pid
#        print "pid = $pid\n";
        $currentCmds{$pid} = $cmdID;
        ++$numCmds;
    } else {
        die "EvCmdMgr::doCmdExecution=> Cannot fork: $!" unless defined $pid;
        #child
        exec($cmdStr);
    }
}
        
#
# doSynchNotify ($cmdID, $cmdStr) 
# 
# Execute the notification as a synchronous lib call.
# cmdStr is a reference to a hash containing all the parameters to eNotify.
#
sub doSynchNotify ($$)
{
    my ($cmdID, $cmdStr) = @_;

    # execute notification
    my $status = eNotify($cmdStr); 

    # save results
    push(@cmdStatus, $cmdID, $status);
}
    

#
# getCmdStatus()
#
# Returns a list of (cmdID, exitSTatus) pairs for all commands that 
# have terminated since the last call to this subroutine.
#
sub getCmdStatus ()
{
    my @results = @cmdStatus;

    @cmdStatus = ();            # clear global

    my ($pid, $status, $cmdID);


    while ( @deadCmds > 0 ) {   # terminated cmds available?
        # Retrieve the pid & exitStatus for the first available cmd
        ($pid, $status) = shift2(@deadCmds);
        
        # Look up the cmdID for the retrieved cmd.
        $cmdID = $currentCmds{$pid};

	next if (!defined($cmdID));	# Some other child proc.

        # Save the same for return
        push(@results, $cmdID, $status);

        # Reap
        delete $currentCmds{$pid};
    }

    return @results;
}

#
# executePendingCmds()
#
# Initiate the execution of cmds that were initially deferred.
#
sub executePendingCmds ()
{
    while (@pendingCmds > 0) {
        last if ($numCmds >= $maxCmds);
        my ($cmdID, $cmdStr) = shift2(@pendingCmds);
        doCmdExecution($cmdID, $cmdStr);
    }
}

#
# setMaxCmds(maxCmds)
#
# Class method to set the max no. of cmds that can be run concurrently.
#
sub setMaxCmds ($)
{ # Class method
    die "EvCmdMgr::setMaxCmds => Numeric arg required!"
        unless ( scalar(@_) == 1 && isNumeric($_[0]) );
    $maxCmds = shift;
}

1;
