# The EvFilter module provides the functionality for representing and 
# processing event filter objects.  
#


package EvFilter;

use strict;
use Exporter;
use ArrayTools;
use StandardTools;
use HashTools;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT = qw( &match &getNotifyStr );

use vars qw(%notifyFields);

# evProcdName eNotifyOption eNotifyName
my @requiredFields = qw(status application);
my @allFields =      qw(status application pattern);

my $notifyCmd = "eNotify";

#
# new (filterName filterDB)
#
# This constructs and returns a new EvFilter object encapsulating 
# the specification of filter filterName, as described in the filterDB.
#
sub new ($$$)
{
	# class, filterName, filterDB
	#
	die "Usage: EvFilter->new( \$filterName, \$filterDB )"
		unless ( scalar(@_) == 3 );

	my ($class, $filterName, $filterDB) = @_;

        my ($self) = bless {name => $filterName, 
                            db => $filterDB,
                            valid => 0
                            }, $class;

        $self->init();

	return $self;
}

#
# init ()
# Initialize the contents of THIS
#
sub init ($)
{
    my ($this) = shift;
    my ($baseStr) = "eps" . $this->{"name"} . "."; # eps.filterName.
    my $key;
    my $found = 0;
    my $notFound = 0;

    my @fields = qw(pattern status application);

    # Retrieve required fields
    ($found, $notFound) = $this->extractFields(@fields);
    if ( $notFound) {       # Some fields not specified?
        # Some, but not all of the reqd fields are specified
        # Dump this filter. Note that this will override a 
        #previous valid declaration of this filter, but that's okay
        $this->{"valid"} = 0;

        # Print error message
        my $msg = "EvFilter: Discarding filter: "  .  $this->{"name"} 
            . "\n\tAll required fields ( @fields ) not specified\n";
        if ($debug) {
            print ERRLOG $msg, "\n";
        }
        warn($msg);
    } else {
	$this->{"valid"} = 1;
    }
}


# 
# extractFields (this, @fields)
#
# Extracts specified fields from the config file. 
# Returns a count of number of fields found and notFound.
#
sub extractFields ($@)
{ 
# this, fields
    my ($this) = shift;
    my (@fields) = @_;
    my ($found) = 0;
    my ($notFound) = 0;
    my ($baseStr) = "eps." . $this->{"name"} . "."; # eps.filterName.
    my ($db) = $this->{"db"};


    foreach my $field (@fields) {
        # Extract the field
        my $val = $db->getResource($baseStr . $field, 
                                   "Eps.Filter.Field", 0);

        if ($val) {             # val defaults to 0
            #Add the field to THIS
            $this->{$field} = $val;
            $found++;
        } else {
            $notFound++;
        }
    }
    return $found, $notFound;
}


#
# match ($eventData)
#
# Returns 1 if the eventData is matched by this filter pattern, 0 otherwise
#
sub match ($$)
{
    die "EvFilter::match => Usage: filter->match(\$event)"
        unless ( scalar(@_) == 2 );
    my ($this, $eventData) = @_;
#    print "filter: ",  $this->{"name"}, " event= ", $eventData, "\n";
    if ($eventData =~ /$this->{"pattern"}/) {
        if ($debug) {
            print ERRLOG "EvFilter::match => Matched filter: ", $this->{"name"}, "\n";
        }
        return 1;
    } else {
        return 0;
    }
}

#
# getNotifyStr($event)
#
# Construct an appropriate notify command for the event, 
# using this filter spec.
#
sub getNotifyStr ($$)
{
    die "EvFilter::getNotifyStr => Usage: filter->getNotifyStr(\$event)"
        unless ( scalar(@_) == 2 );

    my ($this, $eventData) = @_;

    # Construct the notify string
    my %notifyHash;
    my ($str, $key, $value);

    # Initial value for eventTime, converter might override.
    $notifyHash{eventTime} = eNotifyGuts::extractTime($eventData);

    for (@requiredFields) {
	# Insure all required fields are present and insert them
	# into the hash structure.

	# Get our string value
        $str = $this->{$_};

	if (!defined($str)) {
		warn "$0 - getNotifyStr: required field ($_) missing from filter definition\n";
		return;
	}

        $notifyHash{$_} = $str;
    }
    # Arguments all processed.

    # Message detail, append event data
    my @detail;
    push (@detail, $eventData);
    $notifyHash{detail} = \@detail;
    if ($debug) {
        print ERRLOG "notifyHash: \n";
        my $tmp = \%notifyHash;
        printHash(*STDOUT, $tmp);
    }
    return \%notifyHash;
}

1;
