# The EvMonitor module is responsible for providing an event input stream,
# sourced off the system log.
#
package EvMonitor;


use strict;
use Exporter;
use Time::Local;
use IO::Select;
use IO::File;
use StandardTools;

use vars qw(@ISA @EXPORT @EXPORT_OK $eventStream $select $initFlag $eventLog $lastTime @months );

@ISA = qw(Exporter);
@EXPORT = qw(&getEvent &initEventStream &setLastEvent &setLastTime &setEventLogName);

my ($eventStream, $select);

my $initFlag = 0;
my $eventLog = "/var/adm/syslog";
my $lastEvent = "";
my $lastTime = 0;          # in secs since EPOCH
my @months = ("jan", "feb", "mar", "apr", "may", "jun", 
              "jul", "aug", "sep", "oct", "nov", "dec");
my $sleepTime = 1;


# getEvent: Return a new event from the input stream.
# optional args: timeOutInterval -- Return if no event obtained 
# within specified interval.

sub getEvent ($)
{
    my $timeOut = undef;

    $timeOut = shift if ( scalar(@_) == 1 );

    # Verify that the fileHandle is still good.
    if (!$initFlag or ( (stat($eventLog))[3] == 0 ) ) {
        # Re-init the filehandle
        initEventStream();
    }

    my $event = 0;
    my $doneFlag = 0;
    my $sleepCount = 0;

    while (!$doneFlag) {
        
        while ($event = $eventStream->getline()) { # Retrieve next event
            chomp($event);
            $doneFlag = acceptableEventp($event);
            last if $doneFlag;  # Terminate loop if event is acceptable.
        }

        # No more pending events, effectively do a tail -f on the 
        # input event log 
        if (!$doneFlag) {
            $eventStream->clearerr();
            $sleepCount += $sleepTime;
            if ($sleepCount >= $timeOut) {
                $doneFlag = 1;
            } else {
                sleep $sleepTime;
            }
            
        }
    }


    if ($event) {
        return $event;
    } else {
        return;
    }
}

#
# initEventStream: Initialize the input event stream.
#
sub initEventStream ()
{
    undef $eventStream;       # close existing handle

    $eventStream = new IO::File;  
    $eventStream->open($eventLog) 
        or die "Could not open eventlog $eventLog: $!";

    $initFlag = 1;
    
}

# 
# acceptableEventp (event)
# 
# Returns 1 if the event is acceptable for further processing, 0 if not.
# The acceptability criterion is that the event should be newer than the 
# last event seen/processed, as captured in $lastTime
#
sub acceptableEventp ($)
{
    my $event = shift;


    # Compare timestamp with the stamp of last processed event. 
    # This could be a init value (upon startup) or a value read from the 
    # evProcd log/audit file (upon restart)

    # Also consider refining the match to the entire event.
    
    my $newTime = extractTime($event);
    if ( ($newTime >= $lastTime) && ($event ne $lastEvent) ) {
	print ERRLOG "acceptableEventp: Accepting event\n" if ($debug);
    	return 1;
    } elsif ($newTime < $lastTime) {
	print ERRLOG "acceptableEventp: Rejecting event: earlier than prior event\n" if ($debug);
    } else {
	print ERRLOG "acceptableEventp: Rejecting event: identical to prior event\n" if ($debug);
    }

    return 0;               # reject
}

#
# sub extractTime ($event)
# 
# Extracts the date from the event and returns the same as an integer 
# representing seconds elapsed since the epoch.
# Example event:
#Sep 24 23:12:12 tncat42:2 unix: node 2 missed 4 ticks
#
sub extractTime ($)
{
    my $event = shift;
    
    my ($mon, $day, $hh, $mm, $ss);

    if ($event =~ /(^[A-Za-z]+)\s+(\d+)\s+(\d+):(\d+):(\d+)/) {
        ($mon, $day, $hh, $mm, $ss) = ($1, $2, $3, $4, $5);

         # convert month to digit
        my $i = 0;
        foreach my $monthKey (@months) {
            if ($monthKey =~ /$mon/i) {
                $mon = $i;
                last;
            }
            $i++;
        }
	my ($currentDay, $currentMon, $currentYear) = (localtime)[3,4,5];
	my $yr = $currentYear;
	if (($mon > $currentMon) || ($mon == $currentMon && $day > $currentDay))
	{
		# If event month/day is in the future, 
		# assume it's really last year
		$yr--;
	}

       	return timelocal($ss, $mm, $hh, $day, $mon, $yr); # epoch

    } else {
	print ERRLOG "extractTime: Bad event fmt, doesn't match regex [$event]\n"
		if ($debug);
        return 0;
    }

}

#
# setLastEvent (eventData)
# 
# Store eventData as the last event that was successfully processed.
# Also record the time of the event as the lastTime
#
sub setLastEvent ($)
{ 
    $lastEvent = shift;
    setLastTime($lastEvent);
}

#
# setLastTime (eventData | timeInSecs)
# 
# Set the time stamp to be used for determining the acceptability of 
# new events.
sub setLastTime ($)
{
    # arg could be either eventData or time in secs
    my $arg = shift;

    $arg = extractTime($arg) 
        unless isNumeric($arg);

    $lastTime = $arg;          # epoch secs
#    print "lastTime= $lastTime\n";
}

#
# setEventLogName (logName)
# 
# Set the name of the input event file for EPS
#
sub setEventLogName ($)
{
    $eventLog = shift;
}

#autoflush STDOUT 1;
1;
