package Event;

use strict;

use IO::File;
use Exporter;


use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);
@EXPORT = qw( &getEventData &doNotifyPrep &doCmdPrep &cmdComplete 
              &initEventLog &setAuditLogName ); 


my $auditLogName = "/var/adm/epsLog";
my $auditLog;
my $initFlag = 0;

BEGIN 
{
    my $eventID = 0;
    my $cmdID = 0;

    sub newEventID () 
    {
        return ++$eventID;
    }

sub newCmdID ()
{
    return ++$cmdID;
}

}


#
# initEventLog
# Class method for initializing the file handle for the EPS audit log.
#
sub initEventLog ()
{
    # Not a method
    undef $auditLog;            # close if previously open

    $auditLog = new IO::File;  
    $auditLog->open(">> $auditLogName") 
        or die "Could not open auditLog $auditLogName: $!";
    $initFlag = 1;
}

#
# new (Class,  eventData, filter)
# constructs and returns a new Event object 
#
sub new ($$$)
{
    # class, eventData, filter
    #
    die "Usage: Event->new( \$eventData \$filter )"
        unless ( scalar(@_) == 3 );

    my ($class, $eventData, $filter) = @_;

    my ($self) = bless {id => newEventID(), 
                        data => $eventData,
                        filter => $filter,
                        notifyFlag => 0,
                        notifyStatus => "NA",
                        cmdFlag => 0,
                        cmdStatus => "NA",
                        cmdsOutstanding => 0
                        }, $class;

    return $self;
}

#
# getEventData
# Return the raw event message.
#
sub getEventData ($)
{
    my $this = shift;
    return $this->{"data"};
}

#
# doNotifyPrep
# Prepare for the execution of a notify command based on this event.
#
sub doNotifyPrep ($)
{
    my ($this) = shift;
    my $cmdID = newCmdID();
    $this->{"notifyFlag"} = 1;
    $this->{"notifyID"} = $cmdID;
    $this->{"cmdsOutstanding"} = $this->{"cmdsOutstanding"} + 1;

    return $cmdID;
}

#
# doCmdPrep
# Prepare for the execution of a user command based on this event.
#
sub doCmdPrep ($)
{
    my ($this) = shift;
    my $cmdID = newCmdID();
    $this->{"cmdFlag"} = 1;
    $this->{"cmdID"} = $cmdID;
    $this->{"cmdsOutstanding"} = $this->{"cmdsOutstanding"} + 1;

    return $cmdID;
}

#
# cmdComplete(this, cmdID, cmdStatus)
#
# Handle the completion of the cmd identified by cmdID, initially 
# triggered by this event.
#
sub cmdComplete ($$$)
{
    # this, cmdID, status
    die "Usage: Event->cmdComplete( \$cmdID \$status )"
        unless ( scalar(@_) == 3 );

    my ($this, $cmdID, $status) = @_;
    
    my $baseStr = undef;
    
    if ($cmdID == $this->{"notifyID"} ) { # notify cmd?
        $baseStr = "notifyStatus";
    } elsif ($cmdID == $this->{"cmdID"} ) { # arbitrary cmd?
        $baseStr = "cmdStatus";
    } else {
        warn("cmdID not recognized!");
    }

    if (defined $baseStr) {
        # Record the exit status of the cmd.
        $this->{$baseStr} = $status;
        # Take it off the queue
        $this->{"cmdsOutstanding"} = $this->{"cmdsOutstanding"} - 1;
    }

    if ($this->{"cmdsOutstanding"} <= 0) { #Done processing all cmds?
        # Record an audit trail for this event
        $this->logCompletion();
    }
}

#
# logCompletion
#
# Records an audit trail for EPS actions triggered by this event. 
# The audit event is in the following format:
# <logTime>; eventFilter=<filterName>; severity=<severity>; notifyStatus=<notifyStatus>; cmdStatus=<cmdStatus>; event=<rawEvent>;
#
# Note: severity=X, now determined based on event and filter defn at runtime
#
sub logCompletion ()
{
    my $this = shift;

    initEventLog() unless $initFlag;

    # time, filterName, severity, nStatus, cStatus, event
    my $space = ";";
    my $filter = $this->{"filter"};
    my $outStr = localtime() . $space 
                        . " eventFilter=" . $filter->{"name"} . $space
                        . " severity=X" . $space
                        . " notifyStatus=" . $this->{"notifyStatus"} . $space
                        . " cmdStatus=" . $this->{"cmdStatus"} . $space
                        . "event=\"" . $this->{"data"}  ."\";\n";

    $auditLog->print($outStr);
    $auditLog->flush()
}

#
# setAuditLogName
# 
# Set the output file name
sub setAuditLogName ($)
{
    # NOT a method --- class fn
    $auditLogName = shift;
}

1;
