#!/bin/perl -w
#
# Internal implementation routines for eNotify
#

package eNotifyGuts;

use strict;
use Time::Local;
use English;
use Exporter;

use XResources;
use HashTools;
use StandardTools;

use eNotifySecure;

use vars qw(@ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter);
@EXPORT_OK = qw(&init);
@EXPORT = qw(&eNotify &extractTime);

use vars qw($xdefaults $msgdir);

my $converterTimeout = 30;


sub init()
{
	# Get program defaults
	#
	$xdefaults = new XResources "Eps";
	die "No default parameters for Eps\n" if ! $xdefaults->valid();

	#### Perform security check ####

	# Get desired UID/GID for files/directories (get* functions return
	# other "thing" in scalar context).
	#
	my $uid = getpwnam 
	   $xdefaults->getResource("eps.fileOwner", "Appl.fileOwner", "root");
	my $gid = getgrnam 
	   $xdefaults->getResource("eps.fileGroup", "Appl.fileGroup", "sys");

	# Store general security info in hash
	my %securityHash = (
		uid =>		defined($uid) ? $uid : 0,
		gid =>		defined($gid) ? $gid : 0,
		mode =>		0600,
		fatal =>	{ uid => 1, gid => 1 },
	);

	# Validate x-resources security
	die "Aborting: security violation\n" 
		if ! secureXresource( $xdefaults, %securityHash );

	# Revise hash for final security check (unsent msgs directory)
	#
	$securityHash{file} =  $xdefaults->getResource(
				"eNotify.unsentMessagesDirectory",
				"Appl.unsentMessagesDirectory",
				"/opt/compaq/cms/eNotify/unsent");
	$securityHash{mode} =  0700;	# directory permissions
	$securityHash{fatal} = {};	# nothing is fatal (only complains)

	# Check permissions on the unsent messages directory (non-fatal)
	#
	die "Aborting: security violation\n" if ! secureFile( %securityHash );


	#### Security check completed successfully (if we make it here) ####

	# Alter path to include directories for our executables
	#
	my $privateDir = $xdefaults->getResource("eps.privateDirectory",
		"Appl.privateDirectory", "/opt/compaq/cms/.bin");
	my $publicDir = $xdefaults->getResource("eps.publicDirectory",
		"Appl.publicDirectory", "/opt/compaq/cms/bin");

	my $path = "$privateDir:$publicDir:" . $ENV{PATH};
	$ENV{PATH} = $path;

	# Directory for unsent messages
	#
	$msgdir = $xdefaults->getResource(
				"eNotify.unsentMessagesDirectory",
				"Appl.unsentMessagesDirectory",
				"/opt/compaq/cms/eNotify/unsent");

}

# eNotify( %messageParameters )
#
# messageParameters should contain:
#	{
#		# Required parameters...
#		destination =>	[list of destinations],
#		event =>	<event which triggered notification>,
#		description =>	<description of event>,
#
#		# It may also contain...
#		application =>	<triggering appl, default: system>,
#		criticality =>	<event severity, default: major>,
#		detail =>	[list of strings comprising detail],
#	}
#		
# Returns TRUE if the message was queued for transmission, FALSE if
# there were any serious (normally fatal) errors.
#
sub eNotify($)
{
	my ($message) = shift;
	local $SIG{"CHLD"} = 'DEFAULT';

	if (defined($message->{application}) && defined($message->{status}))
	{
		# Second usage variation. Get config file params
		# and optionally run converter. May override
		# command line options.
		#
		getEventConfig($message);
	}

	# Insure required parameters are specified
	#
	if (!defined($message->{destination}) ||
	    !defined($message->{event}) ||
	    !defined($message->{description}))
	{
		print STDERR "eNotify: destination, event, and description required\n";
		return $FALSE;
	}

	# Set default parameters
	#
	if (!defined($message->{application})) {
		$message->{application} = "system";
	}

	if (!defined($message->{criticality})) {
		$message->{criticality} = "Major";
	}

	if (!defined($message->{detail})) {
		my @detail = ( "[No detail]" );
		$message->{detail} = \@detail;
	}

	# Escape all special characters in message detail
	for (@{$message->{detail}}) {
		$_ = quotemeta $_;
	}

	# Insure not too many destinations specified
	if (scalar(@{$message->{destination}}) > 26)
	{
		print STDERR "eNotify: WARNING - too many destinations\n";
	}

	$message->{authorizedTime} = time;

	# For every destination specified, create a separate message file.
	#
	my (@destinations) = @{$message->{destination}};
	for (@destinations)
	{
		# Create file to hold message for current destination
		#	(use pipe so we can get both return code and output)
		#
		if (! (open FD, "mktemp $msgdir/$$.XXXXXX |"))
		{
			print STDERR "Failed to invoke mktemp\n";
			return $FALSE;
		};

		my $msgfile = <FD>;
		chomp($msgfile);
		if (! close FD) {
			if (defined($msgfile)) {
				# msgfile will contain error message
				print STDERR "Failed to create message file: [$msgfile] $!\n";
			} else {
				# really bad failure
				print STDERR "Failed to create message file - no further info\n";
			}
			return $FALSE;
		}

		if (-s $msgfile) {
			print STDERR "File $msgfile is not empty\n";
			return $FALSE;
		}


		if (! (open MSG, "> $msgfile" ))
		{
			print STDERR "Unable to open message file [$msgfile]\n";
			return $FALSE;
		};

		# Output the message to the file
		#
		$message->{destination} = $_;
		printHash( \*MSG, $message );

		close MSG;

		print ERRLOG "Message for dest $_ written to $msgfile\n" if ($debug);
	}

	# Need to signal eNotifyd
	#
	signalENotifyd();

	return $TRUE;
}

# signalENotifyd()
#
# Send a wake-up signal (USR1) to eNotifyd so that it will
# check for unsent messages.
#
sub signalENotifyd()
{
	# Get name of daemon process
	my $daemon = $xdefaults->getResource("eNotify.daemon",
			"Appl.daemon", "eNotifyd");

	# Get the PID for the process
	my $daemonPID;
	if (-f "/tmp/$daemon.pid") 
	{
		# Normal operation, get PID from file.
		#
		$daemonPID = `cat /tmp/$daemon.pid`;
		if (! kill 0,$daemonPID)
		{
			# Stale process ID
			undef $daemonPID;
		}
	}

	# Send the signal
	if (defined($daemonPID) && isNumeric($daemonPID))
	{
		print ERRLOG "Signaling daemon, PID: $daemonPID\n" if ($debug);
		kill('USR1', $daemonPID);
	} else {
		print STDERR "$0: Unable to find daemon $daemon\n";
	}
}

sub getEventConfig($)
{
	my ($message) = @_;
	my $application = $message->{application};
	my $status = $message->{status};

	# Get parameters from eventInfoFile (EpsEvents)
	my $eventInfoFile = $xdefaults->getResource("eNotify.eventInfoFile",
					"eNotify.eventInfoFile",
					"EpsEvents");
	if (!defined($eventInfoFile))
	{
		warn "eventInfoFile not properly defined in Eps configuration file\n";
		return;
	}

	my $xeventInfo = new XResources $eventInfoFile;
	if (!defined($xeventInfo) || ! $xeventInfo->valid())
	{
		warn "Empty or invalid configuration file $eventInfoFile\n";
		return;
	}

	my $eventInfo = $xeventInfo->getResource("eNotify.$application.$status",
					"eNotify.Application.Status",
					"{ bogus => 'data' }");
	if (!defined($eventInfo) || defined($eventInfo->{bogus})) 
	{
		print ERRLOG "No information to handle event from application $application, status $status\n" if ($debug);
		return;
	}

	$message->{criticality} = $eventInfo->{criticality} if (exists($eventInfo->{criticality}));

	$message->{event} = $eventInfo->{event} if (exists($eventInfo->{event}));

	$message->{description} = $eventInfo->{description} if (exists($eventInfo->{description}));

	if (exists($eventInfo->{destination})) {
		my @destinations = split(/,/, $eventInfo->{destination});
		$message->{destination} = \@destinations;
	}

	my $convStr;
	if (exists($eventInfo->{converter})) {
		# Retrieve/run converter, merge results
		my $conv = $eventInfo->{converter};
		my $quoted = "";
		for (@{$message->{detail}}) {
			$quoted .= "$_\n";
		}
		$quoted = quotemeta($quoted);
		my $result = 0;

		local $SIG{ALRM} = sub { die "timeout"; };
		local $SIG{"CHLD"} = 'DEFAULT';

		eval {
			alarm($converterTimeout); # Insure finite wait
			$convStr = `echo $quoted | $conv`;
			$result = $? >> 8;
			alarm(0);
		};

		if ($@ && $@ =~ /timeout/) {
			$convStr = undef;
			my $msg = "Converter timed out for " .
				"event $message->{event}, " .
				"status $status " .
				"-- Proceeding with default values.";
			warn ($msg);
			return;
		}

		# If converter exits with status 2, reject
		# (just return for now)
		if ($result == 2) {
			print ERRLOG "Event rejected by converter [$conv], event detail [$message->{detail}]\n" if ($debug);
			return;
		}
	}

	my $tmpdb = undef;
	if ($convStr) {		# Got a good converter string?
		# create a new resource DB from the string.
		#
		$tmpdb = new XResources $convStr;
		if (! $tmpdb->valid()) {
			my $msg = "Could not create string DB, appl: " .
				"$application, status $status " .
				"-- Proceeding with default values.";
			warn($msg);
			$tmpdb = undef;
		}
	}

	return if (!defined($tmpdb));

	# Merge the results
	my ($key, $value);
	while (($key, $value) = each(%$message))
	{
		my $newvalue = $tmpdb->getResource($key, $key, "bogus");
		if ($newvalue ne "bogus") {
			$message->{$key} = $newvalue if (defined($newvalue));
		} else {
			$message->{$key} = $value if (defined($value));
		}
	}
}

my @months = ("jan", "feb", "mar", "apr", "may", "jun", 
              "jul", "aug", "sep", "oct", "nov", "dec");

#
# sub extractTime ($event)
# 
# (Stolen from retired module EvMonitor.pm)
#
# Extracts the date from the event and returns the same as an integer 
# representing seconds elapsed since the epoch.
# Example event:
#Sep 24 23:12:12 tncat42:2 unix: node 2 missed 4 ticks
#
sub extractTime ($)
{
    my $event = shift;
    
    my ($mon, $day, $hh, $mm, $ss);

    if ($event =~ /(^[A-Za-z]+)\s+(\d+)\s+(\d+):(\d+):(\d+)/) {
        ($mon, $day, $hh, $mm, $ss) = ($1, $2, $3, $4, $5);

         # convert month to digit
        my $i = 0;
        foreach my $monthKey (@months) {
            if ($monthKey =~ /$mon/i) {
                $mon = $i;
                last;
            }
            $i++;
        }
	my ($currentDay, $currentMon, $currentYear) = (localtime)[3,4,5];
	my $yr = $currentYear;
	if (($mon > $currentMon) || ($mon == $currentMon && $day > $currentDay))
	{
		# If event month/day is in the future, 
		# assume it's really last year
		$yr--;
	}

       	return timelocal($ss, $mm, $hh, $day, $mon, $yr); # epoch

    } else {
	print ERRLOG "extractTime: Bad event fmt, doesn't match regex [$event]\n"
		if ($debug);
        return 0;
    }

}

BEGIN { init(); }

1;
