#!/bin/perl -w
#
# eNotifySecure - security validation routines needed by both
# eNotify and eNotifyd.
#
# Not intended for outside consumption.
#

package eNotifySecure;

use strict;
use Exporter;
use XResources;
use StandardTools;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT = qw(&secureFile &secureXresource);


# secureFile( %desired )
#
# Checks security for a file or directory to insure it matches the
# desired parameters. The parameter "%desired" is a hash which
# will contain the desired options. It may include:
#	file	The file/directory path to check
#	uid	The UID for the owner of the file/directory
#	gid	The GID for the group of the file/directory
#	mode	The desired mode (permissions)
#	fatal	{ [uid => x], [gid => x], [mode => x] }
# If fatal defines any value for uid, gid, or mode, failure of that
# test will produce an error message and cause secureFile to exit
# with the value of FALSE. If not defined and the test fails, a
# warning message will be printed to stderr.
#
# Return value: FALSE if any test fails (fatally), TRUE otherwise.
#
sub secureFile(%)
{
	my %desired = @_;

	# get permissions and ownership for file/directory
	#
	my ($mode, $uid, $gid) = (stat($desired{file}))[2,4,5];

	if ($desired{uid} != $uid)
	{
		my $usr = getpwuid($desired{uid});
		printf STDERR "%s - %s not owned by %s\n",
			defined($desired{fatal}{uid}) ? "ERROR" : "WARNING",
			$desired{file}, 
			defined($usr) ? $usr : "[unknown]";
		return $FALSE if defined($desired{fatal}{uid});
	}
	if ($desired{gid} != $gid)
	{
		my $gr = getgrgid($desired{gid});
		printf STDERR "%s - %s not owned by group %s\n",
			defined($desired{fatal}{gid}) ? "ERROR" : "WARNING",
			$desired{file}, 
			defined($gr) ? $gr : "[unknown]";
		return $FALSE if defined($desired{fatal}{gid});
	}
	if ($desired{mode} != ($mode & 0777))
	{
		printf STDERR "%s - %s has permissions %o, not %o\n",
			defined($desired{fatal}{mode}) ? "ERROR" : "WARNING",
			$desired{file}, 
			$mode, 
			$desired{mode};
		return $FALSE if defined($desired{fatal}{mode});
	}
	return $TRUE;
}

# secureXresource
#
# (convenience function)
# Validates the security parameters for all expected X resource file 
# types (master, local, and user).
# Alters the file parameter of the security hash as needed, then calls 
# secureFile and handles the results.
#
# Return value: Returns FALSE if any test fails (fatally), TRUE otherwise.
#
sub secureXresource($%)
{
	my ($xres) = shift;
	my (%securityHash) = @_;

	for my $filetype ("master", "local", "user")
	{
		$securityHash{file} = $xres->getResource("xrsrc.$filetype.file",
					"Internal.FileType.file", "none");

		if ($securityHash{file} ne "none")
		{
			return $FALSE if (! secureFile( %securityHash ));
		}

	}

	# Everything must be OK
	return $TRUE;
}


1;
