#!/bin/perl -w
#
# (see "perldoc perlpod" for info on POD formatting)
#

=head1 NAME

ArrayTools - Generic utilities for working with arrays.

=head1 SYNOPSIS

    use ArrayTools;

    areArraysEqual( $arrayRef1, $arrayRef2 );
    setDiff( @array1, @array2);
    uniq( @array );
    shift2( \@array);

=head1 DESCRIPTION

This package provides functions to allow the caller to perform
a number of interesting actions involving arrays.

=cut

package ArrayTools;

use strict;
use English;
use Exporter;

use StandardTools;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT = qw(&areArraysEqual &setDiff &uniq &shift2);

sub areArraysEqual($$)
{
	my ($arrayRef1, $arrayRef2) = @_;
	die "Usage: areArraysEqual( arrayRef1, arrayRef2 );\n"
		if (ref($arrayRef1) !~ /ARRAY/ && ref($arrayRef2) !~ /ARRAY/);

	# Immediate abort if one array contains more elements
	return $FALSE if (scalar(@$arrayRef1) != scalar(@$arrayRef2));

	for (my $index = 0; $index < scalar(@$arrayRef1); $index++)
	{
		my $type1 = isNumeric(${$arrayRef1}[$index]);
		my $type2 = isNumeric(${$arrayRef2}[$index]);
		return $FALSE if ($type1 != $type2);

		if ($type1) {
		    # Numeric
		    return $FALSE 
			if (${$arrayRef1}[$index] != ${$arrayRef2}[$index]);
		} else {
		    # String
		    return $FALSE 
			if (${$arrayRef1}[$index] ne ${$arrayRef2}[$index]);
		}
	}

	# We survived all the checks, must be equal
	return $TRUE;
}

# setDiff ( @a1, @a2 )
#
# Subtract elements in @a2 from @a1 and return the reduced array.
# You must pass array references to this subroutine, since passing
# by value causes both arrays to flatten out to one list of an
# unpredetermined length.
#
# Notes:
# Ordering of elements will be preserved in the result.
#
# Caveats:
# Ordering of elements within the two arrays is irrelevant to
# the result.
#
sub setDiff ( \@ \@ ) 
{
    my @a1 = @{$_[0]};  # the array to be reduced
    my @a2 = @{$_[1]};  # The elements to be removed from @a1

    foreach my $el ( @a2 ) {
            for ( my $i = 0; $i < @a1; $i++ ) {
            if ( $el eq $a1[$i] ) {
                splice @a1, $i, 1;
            }
        }
    }

    return @a1;
}

# uniq ( @a )
#
# Given an array of possibly non-unique elements, return an array
# containing the same elements, with duplicates removed.
#
# Caveats:
# uniq() does not preserve existing ordering of elements.
#
sub uniq ( \@ ) 
{
    my @in = @{$_[0]};  # the array to sort
    my %saw;            # temp storage for unique values

    @saw{@in} = ();     # extract unique values as keys into %saw
    return keys %saw;   # and return the array of keys
}

#
# shift2 (\@array)
#
# A la "shift", but returns the first pair of elements in the array.
# Note that the argument is passed in as a reference and is DESTRUCTIVELY
# modified by this subroutine.
#
sub shift2 (\@)
{
    die "shift2: The array should have atleast 2 elements!" 
        unless scalar( @{$_[0]} ) >= 2;

    return splice(@{$_[0]}, 0, 2);
}


sub testArrayTools()
{
	my $array1 =	[1,2,3,4,5];
	my $array1_1 =	[1,2,3,4,5];
	my $array2 =	[1,2,3,4,5,6];
	my $array3 =	[1,2,3,4,6];

	my $array4 =	["A","B","C","D","E"];
	my $array4_1 =	["A","B","C","D","E"];
	my $array5 =	["A","B","C","D","E","F"];
	my $array6 =	["A","B","C","D","F"];

	my $array7 =	["A","B",1,"D",2];
	my $array8 =	["A","B","1","D","2"];

	################
	# areArraysEqual
	################

	# Test with numeric only
	print areArraysEqual($array1,$array1_1) ? "ok 1" : "not ok 1", "\n";
	print areArraysEqual($array1,$array2) ? "not ok 2" : "ok 2", "\n";
	print areArraysEqual($array1,$array3) ? "not ok 3" : "ok 3", "\n";

	# Test with strings only
	print areArraysEqual($array4,$array4_1) ? "ok 4" : "not ok 4", "\n";
	print areArraysEqual($array4,$array5) ? "not ok 5" : "ok 5", "\n";
	print areArraysEqual($array4,$array6) ? "not ok 6" : "ok 6", "\n";

	# Test with one numeric and one string
	print areArraysEqual($array1,$array4) ? "not ok 7" : "ok 7", "\n";
	print areArraysEqual($array5,$array1) ? "not ok 8" : "ok 8", "\n";

	# Test with to very simiar (but different) mixed arrays
	print areArraysEqual($array7,$array8) ? 
		"not ok 9 (expected, but annoying)" : "ok 9 (surprise!)", "\n";

	#########
	# setDiff
	#########

	# Test set diff tool
	print scalar(setDiff(@$array1, @$array1_1)) == 0 ? 
		"ok 10" : "not ok 10", "\n";

	# Test simple difference
	my @result = setDiff(@$array2, @$array1);
	print $result[0] == 6 ? "ok 11" : "not ok 11", "\n";

	# Test strings vs. numeric strings (behaves normally for perl, see 9).
	@result = setDiff(@$array7, @$array8);
	print scalar(@result) == 0 ? "ok 12" : "not ok 12", "\n";

	# Test set diff w/order preservation
	my $array9 =	["C", "Z", "A", "Y", "B", "X", "C"];
	my $expected =	["C", "Z", "Y", "X", "C"];
	@result = setDiff(@$array9, @$array7);
	print areArraysEqual(\@result, $expected) ? "ok 13" : "not ok 13", "\n";

	######
	# uniq
	######
	@result = uniq(@$array9);
	my $expected2 =	["X", "Y", "Z", "A", "B", "C"];
	print scalar(setDiff(@result, @$expected2)) == 0 ?
		"ok 14" : "not ok 14", "\n";

        ########
        # shift2
        ########
        my @array4_copy = @$array4;
        my $expected3 = ["C","D","E"];
        my ($el1, $el2) = shift2(@array4_copy);
        print areArraysEqual($expected3, \@array4_copy) ?
            "ok 15": "not ok 15", "\n";

}

1;
