#!/bin/perl -w
#
# (see "perldoc perlpod" for info on POD formatting)
#

=head1 NAME

HashTools - Generic utilities for working with hashed data types.

=head1 SYNOPSIS

    use HashTools;

    expandStrings( {$string|$stringListRef}, $hashref [, $hashname] );
    lookupToken( @token, $hashref [, $hashname] );
    formatHash( $hashref [, $indent]);
    printHash( $outputDevice, $hashref [, $indent]);
    loadHash( $path );

=head1 DESCRIPTION

This package provides functions to allow the caller to perform
a number of interesting actions involving hashed data types
(hashes, blessed hashes, and objects).

=cut

package HashTools;

use strict;
use English;
use Exporter;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT = qw(&expandStrings &lookupToken &formatHash &printHash &loadHash);

###################################################################
#
#

=head1 FUNCTIONS

=head2 expandStrings

B<SYNOPSIS>

B<expandStrings( {$string|$stringListRef}, $hashref [, $hashname] );>

B<ARGUMENTS>

=over 8

=item * $string/$stringListRef

The string or array of strings containing tokens to be
expanded.

=item * $hashref

Reference to the hash (or hash-type structure) to be used 
to resolve the tokens in $string or $stringListRef.

=item * [ $hashname ]

(Optional) The name of the hash. If specified, any tokens
to be substituted must contain this string as the first
element.

=back

B<DESCRIPTION>

For every string specified, any tokens of the form:
<token[.token[...]]> will be searched for within the specified
hash and, if found, replaced with the corresponding value.
The angle brackets "<>" are required to delimit the token
but should not exist in the key values of the hash. If the
token is not found, it is left undisturbed in the string.
This will allow subsequent calls with different hashes
to resolve any remaining tokens.

If hashname is specified, the tokens in the string are
expected to start with that name. For example:

    String:   "HOSTNAME:       <SysInfo.Hostname>"
    Hash:     { Hostname    => "mynode", 
	        OtherParams => "OtherValues" }
    Hashname: "SysInfo"

Note: hashname may be omitted entirely for blessed types. If omitted,
the tokens *may* begin with the type name. If specified, they *must*
begin with the specified name (regardless of type).

The token may specify the maximum length of the substituted
string. Any token of the form <token:nn> (where nn is some number)
will cause the substituted string to be limited to at most nn
characters. Longer strings will be automatically truncated.

The token may specify custom formatting for the substituted string.
Tokens of the form <token:formatter-format> will cause the program
I<formatter> to be executed (must be in the default search path).
Formatter will be passed the specified I<format> as the first argument
and the substitution value for I<token> (found in the hash) as the
second argument. The value returned by formatter will be substituted
in place of token. If formatter cannot be found or does not
execute properly (non-zero exit status), the raw value of the 
substitution string will be used in place of token.

B<RETURN VALUE>

The modified string or array of strings.

=cut

sub expandStrings
{
	my ($toExpand, $hash, $hashname) = @_;

	# This is the only real check we can make thanks to the
	# variable parameter list. Fortunately, it catches cases
	# where first argument passed is an array rather than
	# a reference (actual 2nd parameter will be a scalar in 
	# that case since the array will be expanded).
	#
	die "ERROR: Usage: expandStrings({string|arrayRef}, hash, [hashname]);\n"
	 	if (ref($hash) =~ /^$/);

	# If first element is actually an array reference, expand
	# each string in that array.
	#
	if (ref($toExpand) =~ /ARRAY/)
	{
		foreach my $line (@$toExpand)
		{
			# Could make recursive call, but an arbitrarily
			# nested string table just doesn't make much
			# sense.
			#
			$line = _expandString($line,$hash,$hashname);
		}
	} else {
		$toExpand = _expandString($toExpand,$hash,$hashname);
	}

	return $toExpand;
}

###################################################################
#
#


# _expandString( singleString, hash, [hashname] )
#
# Performs the functionality described above for expandStrings, only
# it operates on a single string. Not exported - expandStrings provides
# the equivalent interface.
#
sub _expandString
{
	my ($line, $hash, $hashname) = @_;
	my ($finalString);

	# Split the line in question based on < and > for
	# delimiters. Parentheses cause delimiters to be
	# retained in the final array.
	#
	# E.g. "Now is the time for all good <people> to come..."
	# Would yield:
	#	"Now is the time for all good "
	#	"<"
	#	"people"
	#	">"
	#	" to come..."
	#
	my (@components) = split /([<>])/, $line;
	if (scalar(@components) <= 1)
	{
		return $line;
	}

	# contained one or more strings to expand
	my $i = 0;
	while ($i <= $#components)
	{
		if ($components[$i] eq "<" && $components[$i + 2] eq ">")
		{
			# Next element is a token, attempt to find
			#

			# Check for length (token suffixed with :nn) or
			# custom formatter (suffixed with :formatter-format).
			#
			my $length;
			my ($formatter, $format, $token);
			my $fullToken = $components[$i + 1];
			if ($fullToken =~ /^([^:]+):(.*)/)
			{
				# Token contains a colon
				$token = $1;
				my $extension = $2;

				if ($extension !~ /\D/)
				{
					# Extension is numeric, use as
					# max field length
					#
					$length = $extension;
				} elsif ($extension =~ /-/) {
					# Extension isn't a length, assume
					# it's a format specification
					# e.g. <formatter>-<format>
					#
					($formatter, $format) = 
						split /-/, $extension;
				} else {
					warn "HashTools: Unrecognized extension for $fullToken\n";
				}
			} else {
				$token = $fullToken;
			}

			my $newValue;

			# Perform lookup of token
			if ($newValue = lookupToken($token, $hash, $hashname))
			{
				# Limit length of token if specified above
				#
				if (defined($length) && 
				    length($newValue) > $length)
				{
					# Length specified and exceeded,
					# truncate value found.
					#
					$newValue = substr($newValue, 0, $length);
				}

				# If custom formatter defined, run it on
				# the specified token.
				#
				if (defined($formatter) && defined($format))
				{
					# Reset signal handler so we can
					# safely check exit status of cmd
					#
					local $SIG{"CHLD"} = 'DEFAULT';

					# Run formatter in a subshell, 
					# catch cmd error output
					#
					my $cmd = "$formatter '$format' '$newValue'";
					$newValue = `/usr/bin/ksh -c "$cmd 2>&1"`;
					my $status = $? >> 8;
					if ($status != 0)
					{
						# Formatter not found, report
						# error message, use original
						# value of token.
						#
						chomp($newValue);
						warn "HashTools: Problem running formatter: $formatter [$newValue]\n";
						$newValue = $token;
					}
				}

				# Append found value to string
				$finalString .= $newValue;

				# Skip delimiters and token
				$i += 3;

				next;
			}
			else
			{
				# Append delimiter to string
				$finalString .= $components[$i];
			}
		}
		else
		{
			# Append value to string
			$finalString .= $components[$i];
		}

		$i++;
	}

	return $finalString;
}

###################################################################
#
#

=head2 lookupToken

B<SYNOPSIS>

B<lookupToken( $token, $hashref [, $hashname] );>

B<ARGUMENTS>

=over 4

=item * $token

The token containing dot-delimited elements which should be
found in the specified hash. Each element corresponds to a
level of nesting within the hash.

=item * $hashref

The hash containing the data that should be used in an attempt
to resolve the token.

=item * [$hashname]

(Optional) The name of the hash. If specified, the token I<must>
contain that name as its first element. If omitted and the
hash is of a blessed type, the token I<may> begin with that 
type name.

=back

B<DESCRIPTION>

The $token is converted into a hash reference, queried against
$hashref, and any value found is returned.

If $hashname is specified and $token begins with that name,
$hashname is stripped from $token before conversion.

Example:

	$comics = { marxbros => { groucho => "Groucho",
				  harpo => "Harpo",
				  chico => "Chico",
				  zeppo => "Zeppo",
				},
		    stooges => { larry => "Larry",
				 moe => "Moe",
				 curly => "Curly",
			       },
		  };

	$comedian = lookupToken("Comics.marxbros.groucho", 
			        $comics, "Comics");

B<RETURN VALUE>

Returns the value to substitute for token if token was
found, else undefined.

B<HISTORY>

Basic logic taken from "get" method of Ldr.pm.

=cut

sub lookupToken
{
	my ($token, $hash, $hashname) = @_;
	my ($query);

	if (defined $hashname)
	{
		# hashname specified by caller, must be present

		if ($token =~ s/^$hashname\.(.*)/$1/)
		{
			# element extracted, do nothing
		} else {
			# hashname not in token, no match
			return;
		}
	} elsif (ref($hash) !~ /HASH/) 
	{
		# Assume we have a blessed type, token might contain name
		$hashname = ref($hash);
		if ($token =~ s/^$hashname\.(.*)/$1/)
		{
			# element extracted, do nothing
		}
	}

	# Convert "." separated fields into braced fields
	# (e.g. "state.admin" into "{state}{admin}")
	#
	$token =~ s/(\w+)\.*/{$1}/g;

	# Combine with hash dereference
	#
	$query = '$hash->' . $token;

	# print "Query string: $query\n";

	# Eval and return the result.
	return eval $query;
}

###################################################################
#
#

=head2 formatHash

B<SYNOPSIS>

B<formatHash( $hash_ref [, $indent]);>

B<ARGUMENTS>

=over 4

=item * $hash_ref

A reference to the hash to be formatted.

=item * [$indent]

(Optional) String of spaces specifying the amount by which to 
indent each line.

=back

B<DESCRIPTION>

Formats $hash_ref in a human-readable, indented 
format. The $indent parameter specifies the initial level of 
indentation. Recursive calls to this routine will increase that 
value appropriately. The keys for each hash will appear in sorted
order.

The output is formatted in such a way that it can be read back
into a program and turned back into a hash (see I<loadHash> below).

Example:
    @formatted = formatHash( $comics );
 
    print "@formatted";

    Output:
       {
       current => [
	  "Eddie Murphy",
	  "Martin Short",
	  "Robin Williams",
          ],
       marxbros =>    {
          chico => "Chico",
          groucho => "Groucho",
          harpo => "Harpo",
          zeppo => "Zeppo",
          },
       stooges =>    {
          curly => "Curly",
          larry => "Larry",
          moe => "Moe",
          },
       },

B<RETURN VALUE>

Scalar context: A string, with embedded newlines, representing the
formatted hash.

Array context: Array of formatted strings that comprise the hash, newlines
removed.

B<CAVEATS>

This routine may have difficulty formatting nested blessed types
(i.e. objects contained within other objects). The outer type
will format appropriately, but the inner type may not.

B<HISTORY>

Was originally printHash. Formatting moved to this function
for better reuse potential. Printing function printHash now
calls this function, then prints as usual.

=cut

sub formatHash
{
	die "Usage: formatHash( hash [, indent])\n" if ( scalar( @_ ) < 1 );
	my ($hashToPrint) = shift;
	my ($indent);
	my $formatted = "";

	if ( @_ == 1 )
	{
		$indent = shift;
		$indent .= "   ";
	}
	else
	{
		$indent = "   ";
	}

	my ($key, $value);
	$formatted .= sprintf "%s{\n", $indent;
	for $key (sort keys %$hashToPrint)
	{
		$value = $hashToPrint->{$key};
		$formatted .= sprintf "%s%s =>", $indent, $key;

		if ( ref($value) eq "HASH" )
		{
			$formatted .= formatHash($value, $indent );
		}
		elsif ( ref($value) eq "ARRAY" )
		{
			# Print the array as a collection of quoted
			# strings with trailing commas (i.e. an anon array)
			#
			my $aindent = $indent . "   ";

			$formatted .= sprintf " [\n";
			for (@$value)
			{
				$formatted .= sprintf "%s\"%s\",\n", $aindent, quotemeta($_);
			}
			$formatted .= sprintf "%s],\n", $aindent;
			
		}
		else
		{
			$formatted .= sprintf " \"%s\",\n", quotemeta($value);
		}
	}
	$formatted .= sprintf "%s},\n", $indent;

	return wantarray ? split /\n/, $formatted : $formatted;
}

###################################################################
#
#

=head2 printHash

B<SYNOPSIS>

B<printHash( $output_device, $hash_ref [, $indent]);>

B<ARGUMENTS>

=over 4

=item * $output_device

The device (file handle) to which the output should be printed.

=item * $hash_ref

A reference to the hash to be printed.

=item * [$indent]

(Optional) String of spaces specifying the amount by which to 
indent each line of output.

=back

B<DESCRIPTION>

Prints $hash_ref to $output_device in a human-readable, indented 
print. The $indent parameter specifies the initial level of 
indentation. The keys for each hash will appear in sorted order.
The function formatHash() is called to perform the actual
formatting.

Example:
    printHash( *STDOUT, $comics );

    Output:
    {
       marxbros =>    {
          chico => Chico
          groucho => Groucho
          harpo => Harpo
          zeppo => Zeppo
       }
       stooges =>    {
          curly => Curly
          larry => Larry
          moe => Moe
       }
    }
B<RETURN VALUE>

None.

B<CAVEATS>

See formatHash().

B<HISTORY>

Was similar to Ldr::printLdr, only with slight differences 
with respect to the printing of braces. Modified to eliminate
those differences.

The formatting capability has been split from this function
to create the formatHash function (thus making formatting
w/out printing available to callers). This function calls
formatHash and then prints the result.

=cut

sub printHash
{
	die "Usage: printHash( dest, hash [, indent])\n" if ( scalar( @_ ) < 2 );
	my ($dev) = shift;
	my ($hashToPrint) = shift;
	my ($indent) = "";

	if ( @_ == 1 )
	{
		$indent = shift;
	}

	my @formattedHash = formatHash($hashToPrint,$indent);

	for (@formattedHash) {
		print $dev "$_\n";
	}
}

###################################################################
#
#

=head2 loadHash

B<SYNOPSIS>

B<printHash( $path );>

B<ARGUMENTS>

=over 4

=item * $path

The full or relative path to the file containing the hash to load.

=back

B<DESCRIPTION>

Loads an anonymous hash from $path and returns a reference to it.
The hash in $path can be created either manually with a text
editor or programatically using I<printHash> or some other routine.

Example:

    $comics = loadHash( "/tmp/comics" );
    printHash( \*STDOUT, $comics) if (defined($comics));

    Output:
    {
       marxbros =>    {
          chico => Chico
          groucho => Groucho
          harpo => Harpo
          zeppo => Zeppo
       }
       stooges =>    {
          curly => Curly
          larry => Larry
          moe => Moe
       }
    }
B<RETURN VALUE>

A reference to the loaded hash (if successful), undef otherwise.

B<CAVEATS>

The file specified is expected to contain only a single hash
(which possibly contains other hashes internally). Only a single
hash reference will be returned by this function.

B<HISTORY>

Existed as independent code in a module in applications/eps but was 
needed for tools/internal/troubleshoot as well. Split the
function out and added it to this module to complement printHash.

=cut

sub loadHash($)
{
	my $path = shift;
	my $hash;

	# Insure that the file specified exists and is non-empty
	if (! -s $path) {
		warn "Hash file: $path missing or empty\n";
		return undef;
	}

	open FD, $path or die "Unable to open hash file: $path\n";
	my (@lines) = grep !/^#|^$/, <FD>;
	close FD;

	chomp(@lines);

	# If struct begins with "{" and ends with "}", then assume a hash
	#
	if ($lines[0]  =~ /^\s*{\s*$/ &&
	    $lines[-1] =~ /^\s*},{0,1}\s*$/)
	{
		# Turn it into a hash reference and return it.
		#
		$hash = eval( join(' ', @lines) );
	} else {
		warn "Unrecognized format in file: $path, hash expected.\n";
	}

	return $hash;
}


1;
