#!/bin/perl -w
#
# (see "perldoc perlpod" for info on POD formatting)
#
# Conventions used in this file/document:
#	LDR (all caps) refers to raw LDR-format data.
#	Ldr (only L upper case) refers to an Ldr object (what this
#	file defines).
#

=head1 NAME

Ldr - Object for creating and manipulating LDR data.

=head1 SYNOPSIS

    use Ldr;

    $myLdr = new Ldr [$raw_LDR_data]
    $myLdr->printLdr($device [, $indent]);

    my $string = $myLdr->toLdr();

    my $token = $myLdr->get(dot.delimited.string);

    $myLdr->set(dot.delimited.string, value);

=head1 DESCRIPTION

This package allows the caller to manipulate an LDR-format
message as an object. Upon creation, the LDR message is converted
to a nested-hash structure. The get and set functions allow
elements to be retrieved from and added to the object. There
is presently no way to remove an element from an Ldr object.

The toLdr function will convert the Ldr object back into raw
LDR data (the ordering of the fields will almost certainly be
different than in the original message, but this is acceptable
by definition). The printLdr function will print a given
Ldr object to the specified device (e.g. *STDOUT).

=cut

package Ldr;

use strict;
use HashTools;

###################################################################
#
#

=head1 FUNCTIONS

=head2 new

B<SYNOPSIS>

B<$myLdr = new Ldr [$rawLdrData];>

B<ARGUMENTS>

=over 8

=item * $rawLdrData

(Optional) A string containing the raw LDR data to be converted 
into an Ldr object. If this data is not provided, an empty
Ldr object will be created.

=back

B<DESCRIPTION>

Creates a new instance of an Ldr object. If $rawLdrData is
provided, that data is converted into a hash and stored within
the object for later retrieval.

B<RETURN VALUE>

The newly created Ldr object if successful, undef otherwise.

=cut
sub new
{
	die "Usage: Ldr::new <ldrdata>\n" if ( $#_ != 1 );

	my $class = shift;
	my $self = shift;
	my ( %ldr );

	# Morph the LDR format data into an anon hash initialization
	#
	# First, convert semicolons into commas and newlines
	# Second, convert "=" into " => "
	# Third, add a newline after every opening brace
	# Forth, add parens around the whole thing.
	#
	$self =~ s/\$/\\\$/g;
	$self =~ s/\;/,\n/g;
	$self =~ s/=/ => /g;
	$self =~ s/{/{\n/g;
	$self = "( $self )";

	# Evaluate the morphed string to get the anon hash itself
	#

	# Disable warnings for remainder of this subroutine
	# (eval gives warning on "systemInfo.system" since "system"
	# has special meaning).
	#
	local ($^W) = 0;

	%ldr = eval "{ $self }";
	$self = \%ldr;

	bless $self, $class;

	return $self;
}

###################################################################
#
#

=head2 printLdr

B<SYNOPSIS>

B<$myLdr-E<gt>printLdr($device [, $indent]);>

B<ARGUMENTS>

=over 4

=item * $device

The device (file handle reference) to print the information to
(e.g. *STDOUT).

=item * $indent

(Optional) The initial level of indentation for printing of the Ldr object.

=back

B<DESCRIPTION>

Prints an Ldr object in an indented, human-readable format to the
device specified. The $indent parameter specifies the initial level 
of indentation. The keys in each hash will be sorted prior to
printing.

B<RETURN VALUE>

None.

B<HISTORY>

Merged original capabilities into printHash and altered code
to call that function.

=cut

# Method to print the LDR object. 
#
sub printLdr
{
	die "Usage: printLdr(device [,indent])\n" if (scalar(@_) < 2);
	my $hashvar = shift;
	my $device = shift;
	my ($indent);

	if ( $#_ eq 0 )
	{
		$indent = shift;
	}
	printHash($device, $hashvar, $indent);
}

###################################################################
#
#

=head2 toLdr

B<SYNOPSIS>

B<$myLdr-E<gt>toLdr(void);>

B<ARGUMENTS>

=over 4

=item None.

=back

B<DESCRIPTION>

Converts the Ldr object into a string containing the raw LDR
data commonly sent between programs or output to files. LDR data
contains only printable ASCII characters.

B<RETURN VALUE>

The string containing the object converted to raw LDR format.

=cut

sub toLdr
{
	my $hashvar = shift;
	my ($key, $value);
	my $returnValue = "";

	for $key (sort keys %$hashvar)
	{
		$value = $hashvar->{$key};
		$returnValue .= $key . "=";
		if (ref($value) eq "HASH")
		{
			# Complex value, convert to string recursively
			#
			$returnValue .= "{" . toLdr($value) . "};";
		}
		else
		{
			# Simple value, convert to string directly
			#
			if ( $value =~ /^\d+$/ )
			{
				# Numeric value, don't quote
				$returnValue .= $value . ";";
			} else
			{
				# Contains alpha characters, quote string
				$returnValue .= "\"" . $value . "\";";
			}
		}
	}

	return $returnValue;
}

###################################################################
#
#

=head2 get

B<SYNOPSIS>

B<$myLdr-E<gt>get( $token );>

B<ARGUMENTS>

=over 4

=item * $token

The token containing dot-delimited elements which should be
found in the Ldr object. Each element corresponds to a
level of nesting within the Ldr object.

=back

B<DESCRIPTION>

HashTools::lookupToken is called to retrieve the specified
value from the hash corresponding to this object.

Example:

    If the Ldr object "$myLdr" contains the following data:

	{ marxbros => { groucho => "Groucho",
			harpo => "Harpo",
			chico => "Chico",
			zeppo => "Zeppo",
		      },
	  stooges =>  { larry => "Larry",
			moe => "Moe",
			curly => "Curly",
		      },
	};

    Then the query:

	$comedian = $myLdr->get( "marxbros.groucho" );

    Will return "Groucho".

B<RETURN VALUE>

Returns the value corresponding to $token if $token was
found, else undefined.

B<HISTORY>

The logic for this routine was originally implemented in "Ldr" itself.
It has since been moved to HashTools::lookupToken for more general
purpose availability.

=cut

sub get
{
	die "Ldr::get: Usage: get <ldr-element>\n" if ( $#_ ne 1 );
	my ($hashvar) = shift;
	my ($desired) = shift;

	return lookupToken($desired, $hashvar);
}

###################################################################
#
#

=head2 set

B<SYNOPSIS>

B<$myLdr-E<gt>set( $token, $value );>

B<ARGUMENTS>

=over 4

=item * $token

The token containing dot-delimited elements which should be
added to or modified in the Ldr object. Each element corresponds to a
level of nesting within the Ldr object.

=item * $value

The value which should be associated with the element designated
by $token.

=back

B<DESCRIPTION>

The contents of $value are added to the Ldr hash in the
location specified by $token. If $token or any of its
components are undefined, they are automatically created.

B<RETURN VALUE>

Returns the results of hash insertion.

=cut

sub set
{
	die "Ldr::set: Usage: set <ldr-element>, <value>\n" if ( $#_ ne 2 );
	my ($hashvar, $element, $newvalue) = @_;
	my ($update);

	# Convert "." separated fields into braced fields
	# (e.g. "state.admin" into "{state}{admin})
	#
	$element =~ s/(\w+)\.*/{$1}/g;

	# Combine the brace-separated fields w/a hash dereference
	$update = '$hashvar->' . $element . " = \"$newvalue\"";

	# print "Desired update string: $update\n";

	# Eval and return the result.
	return eval $update;
}

1;
