#!/bin/perl -w
#
# (see "perldoc perlpod" for info on POD formatting)
#

=head1 NAME

StandardTools - Generic tools and utilities for any program.

=head1 SYNOPSIS

    use StandardTools;

    print ERRLOG "This is a debug message\n" if ($debug);

    print "$value is a number\n" if (isNumeric($value));

=head1 DESCRIPTION

This package provides functions and constants to allow the caller to perform
various standard activities. It includes values for debug (an ERRLOG
file handle to receive debug output and a $debug flag to indicate
whether or not debug is active). It also provides the constants $TRUE
and $FALSE for the user to utilize in his/her programs. Also provided
is a function, isNumeric(), which tests whether it's argument is a
numeric value or not.

=cut

package StandardTools;

use strict;
use English;
use Exporter;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT = qw($TRUE $FALSE $debug *ERRLOG &isNumeric);

# Define these basic constants
use vars qw($TRUE $FALSE);
$TRUE=1;
$FALSE=0;

# Handle debug activation at a global level so all including
# modules will get this info when they include this module.
#
use vars qw($debug);
my ($debug_flag, $debug_file, $perl_debug_flag);

$debug = $FALSE;
if ( defined($debug_flag = $ENV{"TDM_DEBUG"}) )
{
	# We DO NOT want called routines to see this variable!
	#
	delete $ENV{"TDM_DEBUG"};

	# Set separate envmt variable so called routines will
	# have debug active.
	#
	$ENV{"TDM_DEBUG_PERL"} = $debug_flag;
}

# Check the separate envmt variable either just created or set
# by the caller.
#
if ( defined($perl_debug_flag = $ENV{"TDM_DEBUG_PERL"}) ) 
{
	# Can someday include more interesting check of $debug_flag here...
	#
	$debug_flag = $perl_debug_flag;
	if ( uc($debug_flag) eq "ALL" ||
	     uc($debug_flag) eq "PERL" )
	{
		$debug = $TRUE;
	}
}

# Uncomment the following line to override environment variable setting
# and activate debug unconditionally.
#
#$debug = $TRUE;

# If debug is to be activated, make sure the file handle ERRLOG
# is defined and points to either STDERR or to the file specified
# by the environment variable TDM_DEBUG_FILE.
#
# If unable to open ERRLOG, warn and turn off debug.
#
if ( $debug )
{
	my $file_open = 0;
	# print "$0: debug on ($debug_flag), ";

	# See if we should send debug output to a file
	#
	if ( defined($debug_file = $ENV{"TDM_DEBUG_FILE"}) )
	{
		if (! open ERRLOG, ">>$debug_file")
		{
			warn "Unable to open debug log file: $debug_file\n";
		} else {
			# print "redirecting debug output to $debug_file\n";
			$file_open = 1;
		}
	}

	# If an output file wasn't opened, send debug to STDERR
	#
	if ( ! $file_open )
	{
		if ( ! open ERRLOG, ">&STDERR" )
		{ 
			warn "Can't dup STDERR for debug output, deactivating debug\n";
			$debug = $FALSE;
		} else {
			# print "redirecting debug output to STDERR\n";
		}
	}
}

# If debug is still turned on, unbuffer the output stream.
#
if ( $debug )
{
	# make ERRLOG unbuffered
	my $fh = select ERRLOG;
	$| = 1;
	select $fh;

	print ERRLOG "$0: Debug activated ", `date`;
}

# isNumeric($arg)
#
# Returns TRUE if $arg is a numeric value.
#
# Caveats:
#	A numeric string (e.g. "1") will be reported as being
#	numeric even though it's really a string.
#
sub isNumeric($)
{
	my $arg = shift;
	return "$arg" =~ /^\d+$/;
}

1;
