#!/bin/perl -w
#
# (see "perldoc perlpod" for info on POD formatting)
#
# XResources - perl-based access to X resource routines
#

=head1 NAME

XResources - Perl module for accessing X resource database
information.

=head1 SYNOPSIS

  use XResources;

  my ($rsrc) = new XResources "Test";
  my ($rsrc1) = new XResources "String:	Database\n";

  die "No resources" if (! $rsrc->valid());

  my ($filename) = $rsrc->getResource("cms.filename",
				      "App.FileName","junk");

  my (%dbHash) = $rsrc->getDatabase();
  print "Some value: $dbHash{\"some.value\"}\n";

=head1 DESCRIPTION

Provides a simple method to read and access an x-resources
configuration database in a Perl program.

=cut

package XResources;

use strict;
use English;
use Exporter;
use Safe;

use vars qw(@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

use XRsrc;	# Crude hooks to C routines linking to X11 lib routines

# Always handle initialization first (safe to call twice)
#
BEGIN { XRsrc::initialize(); }

# Create a Safe object to handle array/hash initialization.
# Grant it access to the "expr" variable (see "perldoc Safe").
#
my ($safe,$expr);
$safe = new Safe or die "XResources: Can't create a Safe object\n";
$safe->share('$expr');

# When an XResources object is destroyed, insure that the
# corresponding low-level destructor is called to clean-up.
#
sub DESTROY {
	my ($this) = shift;
	XRsrc::destroyDB($this->{db});
	Safe::DESTROY($safe);
}

###################################################################
#
#

=head1 METHODS

=head2 new

B<SYNOPSIS>

B<$myXrsrcs = new XResources $classOrFileName;>
B<$myXrsrc1 = new XResources $databaseString;>

B<ARGUMENTS>

=over 8

=item * $classOrFileName/$databaseString

An application class name (a name beginning with an uppercase 
letter), a file name (containing an optional path component 
and starting with a lower case letter if the path is omitted),
or a string representation of a database (must contain an
embedded newline).

=back

B<DESCRIPTION>

Creates a new database initialized with the contents of the
file or string specified by $classOrFileName. Name may be a class file 
(having an initial uppercase letter and no path specification),
a regular file (either path specification or file name starting
with a lowercase letter), or a string database (containing an
embedded newline).

In the case of a file name, only that information will be
loaded. The name of the file will be stored in the database
as the element xrsrc.master.file.

In the case of s string database, the parameter itself will
be converted to an X resource database.

In the case of a class name, the routine will attempt to
load the file as though it were a standard X resource file
(with minor non-standard extension as noted):

=over 8
=item * Load initial default values [xrsrc.master.file]

    If $XFILESEARCHPATH is defined, look there for <name>

    else, try to load /usr/lib/X11/$LANG/app-defaults/<name>

    else, try to load /usr/lib/X11/app-defaults/<name>

=item * Load site-specific customizations/overrides (non-standard) [xrsrc.local.file]

    If $XLOCALFILESEARCHPATH is defined, look there for <name>

    else, try to load /usr/local/lib/X11/$LANG/app-defaults/<name>

    else, try to load /usr/local/lib/X11/app-defaults/<name>

=item * Load user/site overrides [xrsrc.user.file]

    If $XUSERFILESEARCHPATH is defined, look there for <name>

    else, try to load $XAPPLRESDIR/$LANG/<name>

    else, try to load $XAPPLRESDIR/<name>

    else, try to load $HOME/$LANG/<name>

    else, try to load $HOME/<name>

=back

(Latter X resource loading steps omitted, non-standard)


B<RETURN VALUE>

The newly created XResources database in all cases. See the valid()
method, below.

=cut

sub new
{
	my ($this) = shift;
	my ($ResourceName) = shift;

	my ($returnValue) = bless {
		resource => $ResourceName,
		db => XRsrc::newDB( $ResourceName ),
	}, $this;

	return $returnValue;
}

###################################################################
#
#

=head2 valid

B<SYNOPSIS>

B<$myXrsrcs-E<gt>valid();>

B<ARGUMENTS>

=over 8

=item None.

=back

B<DESCRIPTION>

Determines whether the database is loaded, valid, and usable.
The database object is validated as is the key string
"xrsrc.dbStatus" which is used internally and is set to
either "Valid" or "Invalid". Any value other than "Valid" is
equivalent to "Invalid".

B<RETURN VALUE>

TRUE if the object's database is valid and usable, FALSE otherwise.

=cut

sub valid()
{
	my ($this) = @_;
	return XRsrc::dbValid($this->{db});
}

###################################################################
#
#

=head2 getResource

B<SYNOPSIS>

B<$myXrsrcs-E<gt>getResource($name, $class, $default);>

B<ARGUMENTS>

=over 8

=item * $name

The name specification for the resource to be retrieved.

=item * $class

The class specification for the resource to be retrieved.

=item * $default

The value to return if the query is unsuccessful (will be converted
to an anonymous array or hash if appropriate).

=back

B<DESCRIPTION>

Retrieves specific resources from the XResources object. If a
parameter is found which exactly matches $name, that value will
be returned. If $name is not found, a value which corresponds
to $class will be returned if found. Otherwise, the default
value will be returned.

B<RETURN VALUE>

The resource returned may be a traditional scalar type, 
an anonymous hash, or an anonymous array. If the specified
value is not found in the database, $default will be returned.

B<EXAMPLE>

X resource file contents:

  my.hash:	{hello => "Hello", world => "World"}
  my.array:	["Larry", "Moe", "Curly"]

Calling code:

  my $hashrsrc = $rsrc->getResource("my.hash","Who.Datastruct", "garbage");
  print "$hashrsrc->{hello} $hashrsrc->{world}\n";

  my $arrayrsrc = $rsrc->getResource("my.array","Who.Datastruct", "trash");
  print "${$arrayrsrc}[0], ${$arrayrsrc}[1], and ${arrayrsrc}[2]\n";

Output:

  Hello World
  Larry, Moe, and Curly

B<SEE ALSO>

Refer to documentation on X resource files for further
descriptions of class strings (XrmGetResource(3X11)).

=cut

sub getResource
{
	my ($this) = shift;
	my ($name, $class, $default) = @_;
	my ($rsrc) = XRsrc::getResource($this->{db}, $name, $class, $default);

	# Use additional escape so the following eval doesn't choke
	$rsrc =~ s/"/'/sg;

	# rsrc will always be defined, but may contain default
	#
	if ($rsrc =~ /^[\[\{].*/)
	{
		# Retrieved resource defines an anon array or hash
		#
		# Use "Safe" to interpret this value
		#
		my ($evaled) = $safe->reval('$expr=' . $rsrc);
		if ($@) {
			warn "XResources: Error interpreting array/hash for $name: $rsrc\n";
			$rsrc = $default;
		} else {
			$rsrc = $evaled;
		}
	}

	return $rsrc;
}

###################################################################
#
#

=head2 getDatabase

B<SYNOPSIS>

B<my (%dbHash) = $myXrsrcs-E<gt>getDatabase();>

B<ARGUMENTS>

=over 8

=item None.

=back

B<DESCRIPTION>

Retrieves the contents of the X resource database, converted
to a hash format. The keys of the hash will be the reconstructed
keys for the database, as they are defined in the native
X resource file. The values will be the corresponding values.
If the values in the database specify either an anonymous array
or an anonymous hash, the value will be converted to that type
during the retrieval process.

B<RETURN VALUE>

A hash representing the contents of the X resource database.
The actual database is not altered. If the database cannot be
converted, undef is returned.

=cut

sub
getDatabase()
{
	my ($this) = shift;
	my ($dbString) = XRsrc::listDB($this->{db});

	return if (! defined($dbString));

	my (@dbList) = split(/\n/, $dbString);

	my (%megaHash);

	for (@dbList) {
		# Convert each string into anon hash structure

		my ($key, $value) = split /:/, $_, 2;	# Split on colon

		$key =~ tr/://d;		# Remove colon
		$value =~ s/\s(.*)$/$1/;	# Remove leading white space

		# Convert single quotes to double
		$value =~ s/"/'/sg;

		# Check for anon array/hash
		if ($value =~ /^[\[\{].*/)
		{
			# Anon array or hash, convert to perl type
			#
			my ($evaled) = $safe->reval('$expr=' . $value);
			if ($@) {
				warn "XResources: error interpreting array/hash for $key: $value\n";
			} else {
				$value = $evaled;
			}
		}

		$megaHash{$key} = $value;
	}

	return %megaHash;

}

1;

=head1 AUTHOR

Paul Mickelson, Contractor, Compaq's Tandem Division.

=head1 SEE ALSO

XRsrc.pm, XrmGetResource(3X11).

=cut

