/*
#ident "@(#)$Header: /src/nsc_uw7/usr/src/nsc/uts/nsc/nsc_nodelist.h,v 3.14 2000/03/06 06:09:52 zabby Exp $"
 */
/*                                                
 * Copyright (c) 1995 Locus Computing Corporation 
 *        All Rights Reserved                     
 */                                               

#ifndef _NSC_GENERIC_NSC_NODELIST_H
#define _NSC_GENERIC_NSC_NODELIST_H

/*
 * Macros to work with lists of nodes.
 *
 * These interfaces may be implemented as macros.
 * Care must be taken not to pass any arguments with side effects
 * (such as "foo++") since they may be evaluated multiple times.
 *
 * Available operations are:
 *	CLRALL		Clear all nodes in the nodelist_t.
 *	CLRALL_SET1	Clear all nodes and set one in the nodelist_t.
 *	SETALL		Set all nodes in the nodelist_t.
 *	TESTALL		Return non-zero if nodelist_t not empty
 *	CLR1		Clear an individual node.
 *	SET1		Set an individual node.
 *	TEST1		Test one node; return non-zero if node is present.
 *	CLRN		Clear all nodes in dst which are in src.
 *	SETN		Set all nodes in dst which are in src.
 *	ANDN		Clear all nodes in dst which aren't in src.
 *	FFS		"Find First Set"; return the node number of the first
 *			node, starting from 1; if none, return -1.
 *	FLS		"Find Last Set"; return the node number of the highest
 *			node; if none, return -1.
 *	FFSCLR		"Find First Set and Clear"; return the node number
 *			of the first node, starting from 1, after clearing
 *			that node; if no nodes were set, return -1.
 *	FLSCLR		"Find Last Set and Clear"; return the node number
 *			of the highest set node, starting from 1, after 
 *			clearing that node; if no nodes were set, return -1.
 *	FFC		"Find First Clear"; return the node number of the first
 *			missing node, starting from 1; if none, return -1.
 *	FLC		"Find Last Clear"; return the node number of the 
 *			highest missing node; if none, return -1.
 *	FFCSET		"Find First Clear and Set"; return the node number
 *			of the first missing, starting from 1, after
 *			setting that node; if no nodes were clear, return -1.
 *	FLCSET		"Find Last Clear and Set"; return the node number
 *			of the highest missing node, starting from 1, after
 *			adding that node; if no nodes were missing, return -1.
 *	GET_NEXT	Return the next node in the list (given a cookie)
 */

#if defined(_KERNEL) || defined(_KMEMUSER)
#ifdef NSC_SVR42

#ifdef _KERNEL_HEADERS
#include <nsc/nsc.h>
#include <nsc/nsc_config.h>
#include <util/types.h>
#include <util/sysmacros.h>
#include <util/bitmasks.h>
#else
#include <sys/nsc.h>
#include <sys/nsc_config.h>
#include <sys/types.h>
#include <sys/sysmacros.h>
#include <sys/bitmasks.h>
#endif

/*
 * NOTE:
 * These macros are built on top of base routines that accept an "int".
 * The "nodelist" arguments are of type nodelist_t *.
 * The "node" arguments and return values are not type-casted.
 * However, the "node" arguments should never be negative as this may
 * result in a negative bit shift and corrupting memory at lower addresses.
 */

#define NSC_NODELIST_MAXNODE		NSC_MAX_NODE_VALUE
#define NSC_NODELIST_MINNODE		1
#define NSC_NODELIST_NWORDS \
	BITMASK_NWORDS(NSC_NODELIST_MAXNODE - NSC_NODELIST_MINNODE + 1)

/*
 * The node list structure
 * NOTE:
 *	Some code may depend on this being a structure,
 *	e.g. to do structure assignment.
 * FUTURE/PERF:
 *	Add support to use faster fixed-length bitmask support, if possible.
 *	e.g. if NSC_NODELIST_NWORDS == 1, use BITMASK1_XXX() macros.
 */
typedef struct nsc_nodelist {
	uint_t nl_masks[NSC_NODELIST_NWORDS];
} nsc_nodelist_t;

typedef unsigned long		nsc_nlcookie_t;

/*
 * function prototypes
 */
extern nsc_nodelist_t *nsc_nodelist_alloc(void);
extern void nsc_nodelist_free(nsc_nodelist_t *);
extern int nsc_nodelist_count(nsc_nodelist_t *);
extern nsc_nodelist_t *nsc_nodelist_copy(nsc_nodelist_t *);
extern clusternode_t nsc_nodelist_get_next(nsc_nlcookie_t *, nsc_nodelist_t *);
extern clusternode_t nsc_nodelist_bit2node(int);

#define NSC_NODELIST_CHKNODE(node) \
		((node) >= NSC_NODELIST_MINNODE && \
		 (node) <= NSC_NODELIST_MAXNODE)

#define NSC_NODELIST_BIT2NODE(bit)	nsc_nodelist_bit2node(bit)

/*
 * SET Macros 
 */
#define	NSC_NODELIST_SET1(nlp, node) \
		(ASSERT(NSC_NODELIST_CHKNODE(node)), \
		 BITMASKN_SET1((nlp)->nl_masks, (node) - NSC_NODELIST_MINNODE))

#define	NSC_NODELIST_SETALL(nlp) \
		BITMASKN_SETALL((nlp)->nl_masks, NSC_NODELIST_NWORDS)

#define	NSC_NODELIST_SETN(dstp, srcp) \
		BITMASKN_SETN((dstp)->nl_masks, (srcp)->nl_masks, \
				NSC_NODELIST_NWORDS)

/*
 * TEST Macros
 */
#define	NSC_NODELIST_TEST1(nlp, node) \
		(ASSERT(NSC_NODELIST_CHKNODE(node)), \
		 BITMASKN_TEST1((nlp)->nl_masks, (node) - NSC_NODELIST_MINNODE))

#define	NSC_NODELIST_TESTALL(nlp) \
		BITMASKN_TESTALL((nlp)->nl_masks, NSC_NODELIST_NWORDS)

/* CLEAR Macros
 */
#define	NSC_NODELIST_CLR1(nlp, node) \
		(ASSERT(NSC_NODELIST_CHKNODE(node)), \
		 BITMASKN_CLR1((nlp)->nl_masks, (node) - NSC_NODELIST_MINNODE))

#define	NSC_NODELIST_CLRALL(nlp) \
		BITMASKN_CLRALL((nlp)->nl_masks, NSC_NODELIST_NWORDS)

#define	NSC_NODELIST_CLRN(dstp, srcp) \
		BITMASKN_CLRN((dstp)->nl_masks, (srcp)->nl_masks, \
				NSC_NODELIST_NWORDS)

/*
 * AND Macros
 */
#define	NSC_NODELIST_ANDN(dstp, srcp) \
		BITMASKN_ANDN((dstp)->nl_masks, (srcp)->nl_masks, \
				NSC_NODELIST_NWORDS)
		
/*
 * CLRALL_SET1 - Clear all bits and Set 1 bit.
 */
#define	NSC_NODELIST_CLRALL_SET1(nlp, node) \
		(ASSERT(NSC_NODELIST_CHKNODE(node)), \
		 BITMASKN_CLRALL_SET1((nlp)->nl_masks, (node) - NSC_NODELIST_MINNODE))

/*
 * Find the set/missing first/last node. In some cases set/clear it.
 * See comments at the beginning of the file.
 */
#define	NSC_NODELIST_FFS(nlp) \
	NSC_NODELIST_BIT2NODE(BTTMASKN_FFS((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FLS(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FLS((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FFSCLR(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FFSCLR((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FLSCLR(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FLSCLR((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FFC(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FFC((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FLC(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FLC((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FFCSET(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FFCSET((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

#define	NSC_NODELIST_FLCSET(nlp) \
	NSC_NODELIST_BIT2NODE(BITMASKN_FLCSET((nlp)->nl_masks, \
				NSC_NODELIST_NWORDS))

/*
 * ALLOCATE a nsc_nodelist_t
 *	- The list returned will already be initialized for immediate use.
 */
#define NSC_NODELIST_ALLOC()		nsc_nodelist_alloc()
#define NSC_NODELIST_FREE(nlp)		nsc_nodelist_free(nlp)

/*
 * INITIALIZE a nsc_nodelist_t
 *	- all entries in supplied list are cleared
 *	- currently nothing more is allocated
 */
#define NSC_NODELIST_INIT(nlp)		NSC_NODELIST_CLRALL(nlp)
#define NSC_NODELIST_DEINIT(nlp)	/* nothing to do */

/*
 * MISC Macros
 */
#define NSC_NLCOOKIE_INIT(cp)		(*(cp) = CLUSTERNODE_INVAL);
#define NSC_NODELIST_COPY(nlp)		nsc_nodelist_copy(nlp)
#define NSC_NODELIST_COUNT(nlp)		nsc_nodelist_count(nlp)

/*
 * GET_NEXT - return the next node in the list (given a cookie)
 *	- cookie must be initialized (see NSC_NLCOOKIE_INIT)
 */
#define NSC_NODELIST_GET_NEXT(cookie, nlp) \
		nsc_nodelist_get_next((cookie), (nlp))

#endif /* NSC_SVR42 */
#endif /* _KERNEL || _KMEMUSER */

#endif /* !_NSC_GENERIC_NSC_NODELIST_H */
