#ident "@(#)$Header: /src/nsc_uw7/usr/src/nsc/uts/nsc/nsc_synch.h,v 3.12 2000/03/06 06:09:55 zabby Exp $"

/*
 * Copyright (c) 1994 Locus Computing Corporation
 *        All Rights Reserved
 */

/*
 * This file provides macros to map generic NSC locking interfaces
 * to SVR42 platform specific locking primitives.
 */

#ifndef _NSC_GENERIC_NSC_SYNCH_H
#define _NSC_GENERIC_NSC_SYNCH_H

#ifdef _KERNEL_HEADERS
#include <util/types.h>
#include <util/debug.h>
#include <util/ksynch.h>
#include <mem/kmem.h>
#else  /* !_KERNEL_HEADERS */
#include <sys/types.h>
#include <sys/debug.h>
#include <sys/ksynch.h>
#include <sys/kmem.h>
#endif /* !_KERNEL_HEADERS */


/*
 * The following interfaces are provided for atomically manipulating
 * integer values:
 *
 *	INIT_ATOMIC_INT(&int, value);
 *	INCR_ATOMIC_INT(&int);
 *	DECR_ATOMIC_INT(&int);
 *	value = READ_ATOMIC_INT(&int);
 *	WRITE_ATOMIC_INT(&int, value);
 *
 *	where "int" is of type ATOMIC_INT_T.
 */
typedef atomic_int_t ATOMIC_INT_T;
#define INIT_ATOMIC_INT(_intp, _value) \
	ATOMIC_INT_INIT(_intp, _value)
#define INCR_ATOMIC_INT(_intp) \
	ATOMIC_INT_INCR(_intp)
#define DECR_ATOMIC_INT(_intp) \
	ATOMIC_INT_DECR(_intp)
#define READ_ATOMIC_INT(_intp) \
	ATOMIC_INT_READ(_intp)
#define WRITE_ATOMIC_INT(_intp, _value) \
	ATOMIC_INT_WRITE(_intp, _value)


typedef struct lockdata {
	pl_t		lock_ipl;
	uchar_t		hier;
	lkinfo_t	lkinfo;
} lockdata_t;

/*
 * The DECL_LOCKDATA_T() is platform specific and should only be used
 * in platform specific lock declaration source files.
 */
#define DECL_LOCKDATA_T(_lockdata, _pl, _hier, _name) \
	lockdata_t _lockdata = {_pl, _hier, {_name, 0, {0, 0}}}

#define EXTERN_LOCKDATA_T(_lockdata) \
	extern lockdata_t _lockdata


/*
 * IPL data type and macros for use with spin locks.  The read/write
 * spin lock macros return ipl values when locking.  The read/write
 * try-lock interface returns INVALIPL on failure to acquire the lock.
 *
 * The get/set IPL macros are used to handle the case where spin locks
 * are unlocked in a different order than they were locked.  The
 * following is a common example of their use:
 *
 *	ipl = LOCK_SHR_RW_SPIN_LOCK(&lock1);
 *	........
 *	if (....) {
 *		LOCK_SPIN_LOCK(&lock2);
 *		SET_SPIN_LOCK_SAVED_IPL(&lock2, ipl);
 *		ipl = GET_SPIN_LOCK_IPL(&lock2);
 *	}
 *	........
 *	UNLOCK_SHR_RW_SPIN_LOCK(&lock1, ipl);
 *
 * The call to SET_SPIN_LOCK_SAVED_IPL() sets lock2's saved IPL value
 * so that a subsequent unlock of lock2 will return the IPL level back
 * to the IPL in effect when lock1 was first locked.
 *
 * The GET_SPIN_LOCK_IPL() macro gets the IPL value that lock2 is
 * configured to lock at and this value is then used to unlock lock1
 * at the appropriate IPL level.
 */
typedef pl_t IPL_T;
#define INVALIPL	INVPL
#define GET_SPIN_LOCK_IPL(_lock) \
	(_lock)->lock_ipl
#define GET_SPIN_LOCK_SAVED_IPL(_lock) \
	(_lock)->saved_ipl
#define SET_SPIN_LOCK_SAVED_IPL(_lock, _ipl) \
	(_lock)->saved_ipl = _ipl
#define GET_RW_SPIN_LOCK_IPL(_lock) \
	(_lock)->lock_ipl


/*
 * Spin Locks
 * ----------
 *
 * In NSC, the spin lock primitives are also responsible for performing
 * explicit IPL raising/lowering.  On uniprocessor systems spin locks
 * do not spin, but they still must raise and lower IPL (at least in
 * the case where a spin lock is defined at an IPL level greater than
 * base.
 *
 * The following interfaces are provided:
 *
 *	INIT_SPIN_LOCK(&lock, lock_lkdata)
 *
 *		This macro initializes the specified lock.  The
 *		lock_lkdata parameter is a variable declared and
 *		initialized in a platform specific file using a
 *		platform specific version of DECL_LOCKDATA_T().
 *		Note that this macro can sleep.
 *
 *	INIT_SPIN_LOCK_NOSLEEP(&lock, lock_lkdata)
 *
 *              Same as above, but initialization cannot sleep.
 *
 *	LOCK_SPIN_LOCK(&lock)
 *
 *		This macro locks the spin lock and raises the IPL
 *		level if necessary (specified in the lkdata structure
 *		at initialization time).  
 *
 *	EQUAL_HRCHY_LOCK_SPIN_LOCK(&lock)
 *
 *		This macro is the same as LOCK_SPIN_LOCK() with the
 *		exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *
 *	locked = TRY_LOCK_SPIN_LOCK(&lock)
 *
 *		This macro does not spin, but instead returns non-zero
 *		if the lock was acquired and 0 if unsuccessful.  On
 *		a uniprocessor this macro always returns non-zero.
 *		This call is allowed to violate the lock hierarchy.
 *
 *	ASSERT_LOCKED_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain spin
 *		lock is currently held by the current context.  This
 *		macro will not fail on systems that cannot support this
 *		check.
 *
 *	ASSERT_UNLOCKED_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain spin
 *		lock is not currently held by the current context.  This
 *		macro will not fail on systems that cannot support this
 *		check.
 *
 *	ASSERT_NO_SPIN_LOCKS_HELD()
 *
 *		This is an assert macro that checks that no spin locks
 *		are currently held on the processor.  This macro will
 *		not fail on systems that cannot support this check.
 *
 *	NO_SPIN_LOCKS_HELD()
 *
 *		This macro can be used to trigger debugging code
 *		(i.e. an informative printf) in the case where spin
 *		locks are held when they should not.
 *
 *	UNLOCK_SPIN_LOCK(&lock)
 *
 *		This macro unlocks a held spin lock.
 *
 *	DEINIT_SPIN_LOCK(&lock)
 *
 *		This macro deallocates any data structures that may be
 *		associated with a spin lock.  On some systems this may
 *		be a no-op.
 */
typedef struct {
	lock_t		spinlock;
	pl_t		lock_ipl;
	pl_t		saved_ipl;
} SPIN_LOCK_T;

#define INIT_SPIN_LOCK(_lock, _lockdata) \
	do { \
		LOCK_INIT(&(_lock)->spinlock, _lockdata.hier, \
			  _lockdata.lock_ipl, &_lockdata.lkinfo, KM_SLEEP); \
		(_lock)->lock_ipl = _lockdata.lock_ipl; \
	} while(0)
#define INIT_SPIN_LOCK_NOSLEEP(_lock, _lockdata) \
	do { \
		LOCK_INIT(&(_lock)->spinlock, _lockdata.hier, \
			  _lockdata.lock_ipl, &_lockdata.lkinfo, KM_NOSLEEP); \
		(_lock)->lock_ipl = _lockdata.lock_ipl; \
	} while(0)
#define LOCK_SPIN_LOCK(_lock) \
	(void)((_lock)->saved_ipl \
		 = LOCK(&(_lock)->spinlock, (_lock)->lock_ipl))
#define EQUAL_HRCHY_LOCK_SPIN_LOCK(_lock) \
	(void)((_lock)->saved_ipl \
		 = LOCK_SH(&(_lock)->spinlock, (_lock)->lock_ipl))
extern int try_lock_spin_lock(SPIN_LOCK_T *);
#define TRY_LOCK_SPIN_LOCK(_lock) \
	try_lock_spin_lock(_lock)
#define ASSERT_LOCKED_SPIN_LOCK(_lock) \
	ASSERT(LOCK_OWNED(&(_lock)->spinlock) == B_TRUE)
#define ASSERT_UNLOCKED_SPIN_LOCK(_lock) \
	ASSERT(LOCK_OWNED(&(_lock)->spinlock) == B_FALSE)
#define ASSERT_NO_SPIN_LOCKS_HELD() \
	ASSERT(KS_HOLD0LOCKS())
#define NO_SPIN_LOCKS_HELD() \
	KS_HOLD0LOCKS()			/* Only use to check affirmative case */
#define UNLOCK_SPIN_LOCK(_lock) \
	UNLOCK(&(_lock)->spinlock, (_lock)->saved_ipl)
#define DEINIT_SPIN_LOCK(_lock) \
	LOCK_DEINIT(&(_lock)->spinlock)


/*
 * Read/Write Spin Locks
 * ---------------------
 *
 * Read/Write spin locks allow for a lock to be locked in either shared
 * or exclusive mode.  NSC code must be implemented to not count on a
 * "true" implementation of shared locks (i.e. on some platforms
 * read/write spin locks may be implemented as regular spin locks).
 *
 * Read/Write spin locks are unnecessary for non-preemptible uniprocessor
 * platforms.
 *
 * The following interfaces are provided:
 *
 *	RW_SPIN_LOCK_T(lock)
 *	STATIC_RW_SPIN_LOCK_T(lock)
 *	EXTERN_RW_SPIN_LOCK_T(lock)
 *
 *		Read/Write spin locks are declared via a macro so that
 *		they can be compiled out of the code for non-preemptible
 *		uniprocessors.
 *
 *	INIT_RW_SPIN_LOCK(&lock, lock_lkdata)
 *
 *		This macro initializes the specified lock.  The
 *		lock_lkdata parameter is a variable declared and
 *		initialized in a platform specific file using a
 *		platform specific version of DECL_LOCKDATA_T().
 *		Note that this macro can sleep.
 * 
 *	INIT_RW_SPIN_LOCK_NOSLEEP(&lock, lock_lkdata)
 *
 *              Same as above, but initialization cannot sleep.
 *
 *	LOCK_EXCL_RW_SPIN_LOCK(&lock)
 *
 *		This macro locks the read/write spin lock exclusive and
 *		raises the IPL level if necessary (specified in the lkdata
 *		structure at initialization time).  The value of the
 *		IPL prior to acquiring the lock is returned.
 *
 *	EQUAL_HRCHY_LOCK_EXCL_RW_SPIN_LOCK(&lock)
 *
 *		This macro is the same as LOCK_EXCL_RW_SPIN_LOCK() with
 *		the exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *		The value of the IPL prior to acquiring the lock is
 *		returned.
 *
 *	TRY_LOCK_EXCL_RW_SPIN_LOCK(&lock)
 *
 *		This macro does not spin, but instead returns the
 *		IPL value prior to acquiring the lock exclusive,
 *		and INVALIPL on failure to acquire the lock.  On
 *		a uniprocessor this macro always returns success.
 *		This call is allowed to violate the lock hierarchy.
 *
 *	ASSERT_LOCKED_EXCL_RW_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write spin	lock is currently held exclusive by
 *		the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	ASSERT_UNLOCKED_EXCL_RW_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write spin	lock is not currently held exclusive
 *		by the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	UNLOCK_EXCL_RW_SPIN_LOCK(&lock, ipl)
 *
 *		This macro unlocks a read/write spin lock currently
 *		held exclusive and sets the IPL level to the
 *		specified level.
 *
 *	LOCK_SHR_RW_SPIN_LOCK(&lock)
 *
 *		This macro locks the read/write spin lock shared and
 *		raises the IPL level if necessary (specified in the lkdata
 *		structure at initialization time).  The value of the
 *		IPL prior to acquiring the lock is returned.
 *
 *	EQUAL_HRCHY_LOCK_SHR_RW_SPIN_LOCK(&lock)
 *
 *		This macro is the same as LOCK_SHR_RW_SPIN_LOCK() with
 *		the exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *		The value of the IPL prior to acquiring the lock is
 *		returned.
 *
 *	TRY_LOCK_SHR_RW_SPIN_LOCK(&lock)
 *
 *		This macro does not spin, but instead returns the
 *		IPL value prior to acquiring the lock shared,
 *		and INVALIPL on failure to acquire the lock.  On
 *		a uniprocessor this macro always returns success.
 *		This call is allowed to violate the lock hierarchy.
 *
 *	ASSERT_LOCKED_SHR_RW_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write spin	lock is currently held shared by
 *		the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	ASSERT_UNLOCKED_SHR_RW_SPIN_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write spin	lock is not currently held shared by
 *		the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	UNLOCK_SHR_RW_SPIN_LOCK(&lock, ipl)
 *
 *		This macro unlocks a read/write spin lock currently
 *		held shared and sets the IPL level to the
 *		specified level. 
 *
 *	DEINIT_RW_SPIN_LOCK(&lock)
 *
 *		This macro deallocates any data structures that may be
 *		associated with a read/write spin lock.  On some systems
 *		this may be a no-op.
 */
typedef struct {
	rwlock_t	rwspinlock;
	pl_t		lock_ipl;
} RW_SPIN_LOCK_T;

#define INIT_RW_SPIN_LOCK(_lock, _lockdata) \
	do { \
		RW_INIT(&(_lock)->rwspinlock, _lockdata.hier, \
			  _lockdata.lock_ipl, &_lockdata.lkinfo, KM_SLEEP); \
		(_lock)->lock_ipl = _lockdata.lock_ipl; \
	} while(0)
#define INIT_RW_SPIN_LOCK_NOSLEEP(_lock, _lockdata) \
	do { \
		RW_INIT(&(_lock)->rwspinlock, _lockdata.hier, \
			  _lockdata.lock_ipl, &_lockdata.lkinfo, KM_NOSLEEP); \
		(_lock)->lock_ipl = _lockdata.lock_ipl; \
	} while(0)
#define LOCK_EXCL_RW_SPIN_LOCK(_lock) \
	RW_WRLOCK(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define EQUAL_HRCHY_LOCK_EXCL_RW_SPIN_LOCK(_lock) \
	RW_WRLOCK_SH(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define TRY_LOCK_EXCL_RW_SPIN_LOCK(_lock) \
	RW_TRYWRLOCK(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define ASSERT_LOCKED_EXCL_RW_SPIN_LOCK(_lock) \
	ASSERT(TRY_LOCK_EXCL_RW_SPIN_LOCK(_lock) == PLINVAL)
#define ASSERT_UNLOCKED_EXCL_RW_SPIN_LOCK(_lock)
#define UNLOCK_EXCL_RW_SPIN_LOCK(_lock, _ipl) \
	RW_UNLOCK(&(_lock)->rwspinlock, _ipl)
#define LOCK_SHR_RW_SPIN_LOCK(_lock) \
	RW_RDLOCK(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define EQUAL_HRCHY_LOCK_SHR_RW_SPIN_LOCK(_lock) \
	RW_RDLOCK_SH(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define TRY_LOCK_SHR_RW_SPIN_LOCK(_lock) \
	RW_TRYRDLOCK(&(_lock)->rwspinlock, (_lock)->lock_ipl)
#define ASSERT_LOCKED_SHR_RW_SPIN_LOCK(_lock)
#define ASSERT_UNLOCKED_SHR_RW_SPIN_LOCK(_lock)
#define UNLOCK_SHR_RW_SPIN_LOCK(_lock, _ipl) \
	RW_UNLOCK(&(_lock)->rwspinlock, _ipl)
#define DEINIT_RW_SPIN_LOCK(_lock) \
	RW_DEINIT(&(_lock)->rwspinlock)


/*
 * Fast Spin Locks
 * ---------------
 *
 * Some platforms provide a fast spin lock primitive with restricted
 * semantics for reduced overhead when locking around short sections
 * of code.  While holding a fast spin lock it is illegal to hold
 * another fast spin lock or normal spin lock.  On those platforms
 * that do not provide this interface, fast spin locks are implemented
 * as normal spin locks.  On uniprocessor systems fast spin locks do
 * not spin, but they still must raise and lower IPL (at least in the
 * case where a fast spin lock is defined at an IPL level greater than
 * base).
 *
 * The following interfaces are provided:
 *
 *	INIT_FAST_SPIN_LOCK(&lock, lock_lkdata)
 *
 *		This macro initializes the specified lock.  The
 *		lock_lkdata parameter is a variable declared and
 *		initialized in a platform specific file using a
 *		platform specific version of DECL_LOCKDATA_T().
 *		Note that this macro can sleep.
 *
 *	LOCK_FAST_SPIN_LOCK(&lock)
 *
 *		This macro locks the fast spin lock.
 *
 *	locked = TRY_LOCK_FAST_SPIN_LOCK(&lock)
 *
 *		This macro does not spin, but instead returns non-zero
 *		if the lock was acquired and 0 if unsuccessful.  On
 *		a uniprocessor this macro always returns non-zero.
 *
 *	UNLOCK_FAST_SPIN_LOCK(&lock)
 *
 *		This macro unlocks a held fast spin lock.
 *
 *	DEINIT_FAST_SPIN_LOCK(&lock)
 *
 *		This macro deallocates any data structures that may be
 *		associated with a fast spin lock.  On some systems this
 *		may be a no-op.
 */
typedef fspin_t FAST_SPIN_LOCK_T;
#define INIT_FAST_SPIN_LOCK(_lock, _lock_lkdata) \
	FSPIN_INIT(_lock)
#define LOCK_FAST_SPIN_LOCK(_lock) \
	FSPIN_LOCK(_lock)
#define TRY_LOCK_FAST_SPIN_LOCK(_lock) \
	FSPIN_TRYLOCK(_lock)
#define UNLOCK_FAST_SPIN_LOCK(_lock) \
	FSPIN_UNLOCK(_lock)
#define DEINIT_FAST_SPIN_LOCK(_lock)


/*
 * Sleep Locks
 * -----------
 *
 * Sleep locks provide simple mutual exclusion.  Sleep locks cannot
 * be interrupted by signals.
 *
 * The following interfaces are provided:
 *
 *	INIT_LOCK(&lock, lock_lkdata)
 *
 *		This macro initializes the specified lock.  The
 *		lock_lkdata parameter is a variable declared and
 *		initialized in a platform specific file using a
 *		platform specific version of DECL_LOCKDATA_T().
 *		Note that this macro can sleep.
 *
 *	LOCK_LOCK(&lock)
 *
 *		This macro locks the sleep lock.
 *
 *	EQUAL_HRCHY_LOCK_LOCK(&lock)
 *
 *		This macro is the same as LOCK_LOCK() with the
 *		exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *
 *	TRY_LOCK_LOCK(&lock)
 *
 *		This macro returns a non-zero value if the lock was
 *		acquired and 0 if unsuccessful.  This call is allowed
 *		to violate the lock hierarchy.
 *
 *	LOCK_LOCKOWNED(&lock)
 *
 *		This macro returns a non-zero value if the current
 *		context already has the specified lock held.
 *
 *	ASSERT_LOCKED_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain sleep
 *		lock is currently held by the current context.  This
 *		macro will not fail on systems that cannot support this
 *		check.
 *
 *	ASSERT_UNLOCKED_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain sleep
 *		lock is currently not held by the current context.  This
 *		macro will not fail on systems that cannot support this
 *		check.
 *
 *	UNLOCK_LOCK(&lock)
 *
 *		This macro unlocks a held sleep lock.
 *
 *	DEINIT_LOCK(&lock)
 *
 *		This macro deallocates any data structures that may be
 *		associated with a sleep lock.  On some systems this may
 *		be a no-op.
 */
typedef sleep_t	LOCK_T;
#define INIT_LOCK(_lock, _lockdata) \
	SLEEP_INIT(_lock, _lockdata.hier, &_lockdata.lkinfo, KM_SLEEP)
#define LOCK_LOCK(_lock) \
	SLEEP_LOCK(_lock, PRIMED)
#define EQUAL_HRCHY_LOCK_LOCK(_lock) \
	SLEEP_LOCK(_lock, PRIMED)	/* sleep lock hier not enforced */
#ifdef _LOCKTEST
#define TRY_LOCK_LOCK(_lock) \
	(!SLEEP_LOCKOWNED(_lock) && SLEEP_TRYLOCK(_lock))
#else
#define TRY_LOCK_LOCK(_lock) \
	SLEEP_TRYLOCK(_lock)
#endif
#define	LOCK_LOCKOWNED(_lock)\
	SLEEP_LOCKOWNED(_lock)
#define ASSERT_LOCKED_LOCK(_lock) \
	ASSERT(SLEEP_LOCKAVAIL(_lock) == B_FALSE)
#define ASSERT_UNLOCKED_LOCK(_lock) \
	ASSERT(SLEEP_LOCKAVAIL(_lock) == B_TRUE)
#define UNLOCK_LOCK(_lock) \
	SLEEP_UNLOCK(_lock)
#define DEINIT_LOCK(_lock) \
	SLEEP_DEINIT(_lock)


/*
 * Read/Write Sleep Locks
 * ----------------------
 *
 * Read/Write sleep locks allow for a lock to be locked in either shared
 * or exclusive mode.  Read/write sleep locks cannot be interrupted by
 * signals.
 *
 * The following interfaces are provided:
 *
 *	INIT_RW_LOCK(&lock, lock_lkdata)
 *
 *		This macro initializes the specified lock.  The
 *		lock_lkdata parameter is a variable declared and
 *		initialized in a platform specific file using a
 *		platform specific version of DECL_LOCKDATA_T().
 *		Note that this macro can sleep.
 *
 *	INIT_RW_LOCK_NOSLEEP(&lock, lock_lkdata)
 *
 *              Same as above, but initialization cannot sleep.
 *
 *	LOCK_EXCL_RW_LOCK(&lock)
 *
 *		This macro locks the read/write sleep lock exclusive.
 *
 *	EQUAL_HRCHY_LOCK_EXCL_RW_LOCK(&lock)
 *
 *		This macro is the same as LOCK_EXCL_RW_LOCK() with
 *		the exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *
 *	TRY_LOCK_EXCL_RW_LOCK(&lock)
 *
 *		This macro does not sleep, but instead returns
 *		a non-zero value upon successfully acquiring the
 *		sleep lock exclusive, and 0 on failure.  This call
 *		is allowed to violate the lock hierarchy.
 *
 *	ASSERT_LOCKED_EXCL_RW_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write sleep lock is currently held exclusive by
 *		the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	ASSERT_UNLOCKED_EXCL_RW_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write sleep lock is not currently held exclusive
 *		by the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	UNLOCK_EXCL_RW_LOCK(&lock)
 *
 *		This macro unlocks a read/write sleep lock currently
 *		held exclusive.
 *
 *	LOCK_SHR_RW_LOCK(&lock)
 *
 *		This macro locks the read/write sleep lock shared.
 *
 *	EQUAL_HRCHY_LOCK_SHR_RW_LOCK(&lock)
 *
 *		This macro is the same as LOCK_SHR_RW_LOCK() with
 *		the exception that the locking context may already have
 *		another lock held at the same hierarchy level.  Use
 *		of this macro in such situations is important for
 *		systems that enforce lock hierarchies.  Care should
 *		be excersised in using this function to avoid deadlocks.
 *
 *	TRY_LOCK_SHR_RW_LOCK(&lock)
 *
 *		This macro does not sleep, but instead returns a
 *		non-zero value on acquiring the lock shared,
 *		and 0 on failure.  This call is allowed to
 *		violate the lock hierarchy.
 *
 *	ASSERT_LOCKED_SHR_RW_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write sleep lock is currently held shared by
 *		the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	ASSERT_UNLOCKED_SHR_RW_LOCK(&lock)
 *
 *		This is an assert macro that checks that a certain
 *		read/write sleep lock is not currently held shared
 *		by the current context.  This macro will not fail on
 *		systems that cannot support this check.
 *
 *	UNLOCK_SHR_RW_LOCK(&lock)
 *
 *		This macro unlocks a read/write sleep lock currently
 *		held shared.
 *
 *	DEINIT_RW_LOCK(&lock)
 *
 *		This macro deallocates any data structures that may be
 *		associated with a read/write sleep lock.  On some systems
 *		this may be a no-op.
 */
typedef rwsleep_t RW_LOCK_T;
#define INIT_RW_LOCK(_lock, _lockdata) \
	RWSLEEP_INIT(_lock, _lockdata.hier, &_lockdata.lkinfo, KM_SLEEP)
#define INIT_RW_LOCK_NOSLEEP(_lock, _lockdata) \
	RWSLEEP_INIT(_lock, _lockdata.hier, &_lockdata.lkinfo, KM_NOSLEEP)
#define LOCK_EXCL_RW_LOCK(_lock) \
	RWSLEEP_WRLOCK(_lock, PRIMED)
#define EQUAL_HRCHY_LOCK_EXCL_RW_LOCK(_lock) \
	RWSLEEP_WRLOCK(_lock, PRIMED)	/* sleep lock hier not enforced */
#define TRY_LOCK_EXCL_RW_LOCK(_lock) \
	RWSLEEP_TRYWRLOCK(_lock)
#define ASSERT_LOCKED_EXCL_RW_LOCK(_lock) \
	ASSERT(RWSLEEP_RDAVAIL(_lock) == B_FALSE)
#define ASSERT_UNLOCKED_EXCL_RW_LOCK(_lock) \
	ASSERT(RWSLEEP_RDAVAIL(_lock) == B_TRUE)
#define UNLOCK_EXCL_RW_LOCK(_lock) \
	RWSLEEP_UNLOCK(_lock)
#define LOCK_SHR_RW_LOCK(_lock) \
	RWSLEEP_RDLOCK(_lock, PRIMED)
#define EQUAL_HRCHY_LOCK_SHR_RW_LOCK(_lock) \
	RWSLEEP_RDLOCK(_lock, PRIMED)	/* sleep lock hier not enforced */
#define TRY_LOCK_SHR_RW_LOCK(_lock) \
	RWSLEEP_TRYRDLOCK(_lock)
#define ASSERT_LOCKED_SHR_RW_LOCK(_lock) \
	ASSERT(RWSLEEP_RDAVAIL(_lock) == B_TRUE) /* NSC_XXX: not a real check */
#define ASSERT_UNLOCKED_SHR_RW_LOCK(_lock) \
	ASSERT(RWSLEEP_RDAVAIL(_lock) == B_TRUE) /* NSC_XXX: not a real check */
#define UNLOCK_SHR_RW_LOCK(_lock) \
	RWSLEEP_UNLOCK(_lock)
#define DEINIT_RW_LOCK(_lock) \
	RWSLEEP_DEINIT(_lock)


/*
 * Multithreading Sleep Locks
 * --------------------------
 *
 * Multithreading sleep locks are used to handle locking situations that
 * only arise on preemptible or SMP systems.  On a non-preemptible
 * uniprocessor, these locks compile out.  A subset of the full
 * sleep lock interfaces are provided.
 */
#define MTHREAD_LOCK_T(_lock) \
	LOCK_T _lock
#define INIT_MTHREAD_LOCK(_lock, _lockdata) \
	INIT_LOCK(_lock, _lockdata)
#define LOCK_MTHREAD_LOCK(_lock) \
	LOCK_LOCK(_lock)
#define EQUAL_HRCHY_LOCK_MTHREAD_LOCK(_lock) \
	EQUAL_HRCHY_LOCK_LOCK(_lock)
#define TRY_LOCK_MTHREAD_LOCK(_lock) \
	TRY_LOCK_LOCK(_lock)
#define ASSERT_LOCKED_MTHREAD_LOCK(_lock) \
	ASSERT_LOCKED_LOCK(_lock)
#define ASSERT_UNLOCKED_MTHREAD_LOCK(_lock) \
	ASSERT_UNLOCKED_LOCK(_lock)
#define UNLOCK_MTHREAD_LOCK(_lock) \
	UNLOCK_LOCK(_lock)
#define DEINIT_MTHREAD_LOCK(_lock) \
	DEINIT_LOCK(_lock)
	

/*
 * Multithreading Read/Write Sleep Locks
 * -------------------------------------
 *
 * Multithreading read/write sleep locks are used to handle locking
 * situations that only arise on preemptible or SMP systems.  On a
 * non-preemptible uniprocessor, these locks compile out.  A subset
 * of the full read/write sleep lock interfaces are provided.
 */
#define MTHREAD_RW_LOCK_T(_lock) \
	RW_LOCK_T _lock
#define INIT_RW_MTHREAD_LOCK(_lock, _lockdata) \
	INIT_RW_LOCK(_lock, _lockdata)
#define LOCK_EXCL_RW_MTHREAD_LOCK(_lock) \
	LOCK_EXCL_RW_LOCK(_lock)
#define EQUAL_HRCHY_LOCK_EXCL_RW_MTHREAD_LOCK(_lock) \
	EQUAL_HRCHY_LOCK_EXCL_RW_LOCK(_lock)
#define TRY_LOCK_EXCL_RW_MTHREAD_LOCK(_lock) \
	TRY_LOCK_EXCL_RW_LOCK(_lock)
#define ASSERT_LOCKED_EXCL_RW_MTHREAD_LOCK(_lock) \
	ASSERT_LOCKED_EXCL_RW_LOCK(_lock)
#define ASSERT_UNLOCKED_EXCL_RW_MTHREAD_LOCK(_lock) \
	ASSERT_UNLOCKED_EXCL_RW_LOCK(_lock)
#define UNLOCK_EXCL_RW_MTHREAD_LOCK(_lock) \
	UNLOCK_EXCL_RW_LOCK(_lock)
#define LOCK_SHR_RW_MTHREAD_LOCK(_lock) \
	LOCK_SHR_RW_LOCK(_lock)
#define EQUAL_HRCHY_LOCK_SHR_RW_MTHREAD_LOCK(_lock) \
	EQUAL_HRCHY_LOCK_SHR_RW_LOCK(_lock)
#define TRY_LOCK_SHR_RW_MTHREAD_LOCK(_lock) \
	TRY_LOCK_SHR_RW_LOCK(_lock)
#define ASSERT_LOCKED_SHR_RW_MTHREAD_LOCK(_lock) \
	ASSERT_LOCKED_SHR_RW_LOCK(_lock)
#define ASSERT_UNLOCKED_SHR_RW_MTHREAD_LOCK(_lock) \
	ASSERT_UNLOCKED_SHR_RW_LOCK(_lock)
#define UNLOCK_SHR_RW_MTHREAD_LOCK(_lock) \
	UNLOCK_SHR_RW_LOCK(_lock)
#define DEINIT_RW_MTHREAD_LOCK(_lock) \
	DEINIT_RW_LOCK(_lock)


/*
 * Condition locks
 * ---------------
 *
 * Condition locks are locks associated with a condition variable.
 * These locks may be either spin locks or normal sleep locks,
 * depending on the condition variable primitives supported by the
 * underlying platform.  Locking the condition lock may cause
 * either spinning or sleeping and NSC code should be prepared
 * for either case.  Condition locks compile out on non-preemptible
 * uniprocessor platforms.  A subset of the full spin lock interfaces
 * are provided.
 */
typedef SPIN_LOCK_T COND_LOCK_T;
#define INIT_COND_LOCK(_lock, _lockdata) \
	INIT_SPIN_LOCK(_lock, _lockdata)
#define INIT_COND_LOCK_NOSLEEP(_lock, _lockdata) \
	INIT_SPIN_LOCK_NOSLEEP(_lock, _lockdata)
#define LOCK_COND_LOCK(_lock) \
	LOCK_SPIN_LOCK(_lock)
#define TRY_LOCK_COND_LOCK(_lock) \
	TRY_LOCK_SPIN_LOCK(_lock)
#define ASSERT_LOCKED_COND_LOCK(_lock) \
	ASSERT_LOCKED_SPIN_LOCK(_lock)
#define ASSERT_UNLOCKED_COND_LOCK(_lock) \
	ASSERT_UNLOCKED_SPIN_LOCK(_lock)
#define UNLOCK_COND_LOCK(_lock) \
	UNLOCK_SPIN_LOCK(_lock)
#define DEINIT_COND_LOCK(_lock) \
	DEINIT_SPIN_LOCK(_lock)


/*
 * Non-Interruptable Condition Variable Primitives
 * -----------------------------------------------
 *
 * The following interfaces are provided for condition variables
 * that are never interrupted by POSIX signals:
 *
 *	INIT_CONDITION(&cond)
 *
 *		This macro initializes the specified condition
 *		variable.  Note that this macro can sleep.
 *
 *	WAIT_CONDITION(&cond, &cond_lock)
 *
 *		Sleep waiting for a certain condition to be posted
 *		via either SIGNAL_CONDITION() or BROADCAST_CONDITION().
 *		The associated lock should of type COND_LOCK_T.  The
 *		condition lock is atomically unlocked when the
 *		context goes to sleep, and is relocked upon return.
 *		Note that the condition variable is not necessarily
 *		relocked atomically (some platforms may not support
 *		this).
 *		All code that calls WAIT_CONDITION() must tolerate
 *		premature wakeups.  After waking up, the condition
 *		must be rechecked (see the example below).
 *
 *
 *	WAIT_CONDITION_NORELCK(&cond, &cond_lock)
 *
 *		Sleep waiting for a certain condition to be posted
 *		via either SIGNAL_CONDITION() or BROADCAST_CONDITION().
 *		The associated lock should of type COND_LOCK_T.  The
 *		condition lock is atomically unlocked when the
 *		context goes to sleep.  It is NOT relocked upon return.
 *		All code that calls WAIT_CONDITION_NORELCK() must tolerate
 *		premature wakeups.  After waking up, the condition
 *		must be rechecked (see the example below).
 *
 *
 *	SIGNAL_CONDITION(&cond)
 *
 *		Wake up a single context currently waiting on the
 *		specified condition variable via WAIT_CONDITION() or
 *		WAIT_CONDITION_NORELCK().
 *
 *
 *		SIGNAL_CONDITION() should not be called	with the
 *		condition lock held.  On platforms where condition
 *		locks are sleep locks this could cause spurious
 *		scheduling activity.  This would occur if a context
 *		were to wakeup and then go to sleep again on the
 *		condition lock.
 *
 *	BROADCAST_CONDITION(&cond)
 *
 *		Wake up all contexts currently waiting on the
 *		specified condition variable via WAIT_CONDITION() or
 *		WAIT_CONDITION_NORELCK().
 *
 *
 *		BROADCAST_CONDITION() should not be called with the
 *		condition lock held.  On platforms where condition
 *		locks are sleep locks this could cause spurious
 *		scheduling activity.  This would occur if a context
 *		were to wakeup and then go to sleep again on the
 *		condition lock.
 *
 *	DEINIT_CONDITION(&cond)
 *
 *		This macro deallocates any data structures that may be
 *		associated with the condition variable.  On some systems
 *		this may be a no-op.
 *
 * Examples:
 *
 *	Context waiting on a condition:
 *
 *		LOCK_COND_LOCK(&cond_lock);
 *		while (!condition) {
 *			WAIT_CONDITION(&cond, &cond_lock);
 *				:
 *				:
 *		}
 *			:
 *			:
 *		UNLOCK_COND_LOCK(&cond_lock);
 *
 *	Context signalling condition:
 *
 *		LOCK_COND_LOCK(&cond_lock);
 *		<Set the condition>
 *		UNLOCK_COND_LOCK(&cond_lock);
 *		SIGNAL_CONDITION(&cond);
 */
typedef sv_t CONDITION_T;
#define INIT_CONDITION(_cond) \
	SV_INIT(_cond)
#define WAIT_CONDITION_NORELCK(_cond, _cond_lock) \
	do { \
		SV_WAIT(_cond, PRIMED, &(_cond_lock)->spinlock); \
	} while(0)
#define WAIT_CONDITION(_cond, _cond_lock) \
	do { \
		WAIT_CONDITION_NORELCK(_cond, _cond_lock); \
		LOCK_COND_LOCK(_cond_lock); \
	} while(0)
#define SIGNAL_CONDITION(_cond) \
	SV_SIGNAL(_cond, 0)
#define BROADCAST_CONDITION(_cond) \
	SV_BROADCAST(_cond, 0)
#define DEINIT_CONDITION(_cond)
	

/*
 * Interruptable Condition Variable Primitives
 * -------------------------------------------
 *
 * The following interfaces are provided for condition variables
 * that can potentially be interrupted by POSIX signals:
 *
 *	INIT_INTR_CONDITION(&cond)
 *
 *		This macro initializes the specified condition
 *		variable.  Note that this macro can sleep.
 *
 *	WAIT_INTR_CONDITION(&cond, &cond_lock, is_intr, &was_intr)
 *
 *		Sleep waiting for a certain condition to be posted
 *		or for a signal to interrupt the wait.  The is_intr
 *		parameter is a boolean specifying whether the
 *		wait is interruptible by POSIX signals.  The
 *		was_intr flag specifies whether the wait was
 *		interrupted by POSIX signals (B_TRUE is returned)
 *		or woken up by either SIGNAL_INTR_CONDITION() or
 *		BROADCAST_INTR_CONDITION() (B_FALSE is returned).
 *
 *		The associated lock should of type COND_LOCK_T.  The
 *		condition lock is atomically unlocked when the context
 *		goes to sleep, and is relocked upon return.  Note that
 *		the condition variable is not necessarily relocked
 *		atomically (some platforms may not support this).
 *
 *		All code that calls WAIT_INTR_CONDITION() must tolerate
 *		premature wakeups.  After waking up, the condition
 *		must be rechecked (see the example below).
 *
 *	SIGNAL_INTR_CONDITION(&cond)
 *
 *		Wake up a single context currently waiting on the
 *		specified condition variable via WAIT_INTR_CONDITION().
 *
 *		SIGNAL_INTR_CONDITION() should not be called with the
 *		condition lock held.  On platforms where condition locks
 *		are sleep locks this could cause spurious scheduling
 *		activity.  This would occur if a context were to wakeup
 *		and then go to sleep again on the condition lock.
 *
 *	BROADCAST_INTR_CONDITION(&cond)
 *
 *		Wake up all contexts currently waiting on the
 *		specified condition variable via WAIT_INTR_CONDITION().
 *
 *		BROADCAST_INTR_CONDITION() should not be called	with the
 *		condition lock held.  On platforms where condition locks
 *		are sleep locks this could cause spurious scheduling
 *		activity.  This would occur if a context were to wakeup
 *		and then go to sleep again on the condition lock.
 *
 *	DEINIT_INTR_CONDITION(&cond)
 *
 *		This macro deallocates any data structures that may be
 *		associated with the condition variable.  On some systems
 *		this may be a no-op.
 *
 * Examples:
 *
 *	Context waiting on a condition:
 *
 *		LOCK_COND_LOCK(&cond_lock);
 *		while (!condition) {
 *			WAIT_INTR_CONDITION(&cond, &cond_lock);
 *				:
 *				:
 *		}
 *			:
 *			:
 *		UNLOCK_COND_LOCK(&cond_lock);
 *
 *	Context signalling condition:
 *
 *		LOCK_COND_LOCK(&cond_lock);
 *		<Set the condition>
 *		UNLOCK_COND_LOCK(&cond_lock);
 *		SIGNAL_INTR_CONDITION(&cond);
 */
typedef sv_t INTR_CONDITION_T;
#define INIT_INTR_CONDITION(_cond) \
	SV_INIT(_cond)
#define WAIT_INTR_CONDITION_NORELCK(_cond, _cond_lock, _is_intr, _was_intr) \
	do { \
		if (_is_intr) { \
			*(_was_intr) = (SV_WAIT_SIG(_cond, PRIMED, \
						    &(_cond_lock)->spinlock) \
						? B_FALSE : B_TRUE); \
		} \
		else { \
			SV_WAIT(_cond, PRIMED, &(_cond_lock)->spinlock); \
			*(_was_intr) = B_FALSE; \
		} \
	} while(0)
#define WAIT_INTR_CONDITION(_cond, _cond_lock, _is_intr, _was_intr) \
	do { \
		WAIT_INTR_CONDITION_NORELCK(_cond, _cond_lock, \
					    _is_intr, _was_intr); \
		LOCK_COND_LOCK(_cond_lock); \
	} while(0)
#define SIGNAL_INTR_CONDITION(_cond) \
	SV_SIGNAL(_cond, 0)
#define BROADCAST_INTR_CONDITION(_cond) \
	SV_BROADCAST(_cond, 0)
#define DEINIT_INTR_CONDITION(_cond)


/*
 * Non-Interruptable Event Variable Primitives
 * -------------------------------------------
 *
 * The following interfaces are provided for event variables that are
 * never interrupted by POSIX signals. Events are NOT counted, i.e. if
 * SIGNAL_EVENT() is called multiple times between calls to WAIT_EVENT(),
 * WAIT_EVENT() will only be woken-up once.
 *
 *	INIT_EVENT(&event)
 *
 *		This macro initializes the specified event
 *		variable.  Note that this macro can sleep.
 *
 *	WAIT_EVENT(&event)
 *
 *		Sleep waiting for a certain event to be posted
 *		via either SIGNAL_EVENT() or BROADCAST_EVENT().
 *
 *		If there is already an event queued when WAIT_EVENT()
 *		is called, WAIT_EVENT() will not go to sleep.
 *
 *	SIGNAL_EVENT(&event)
 *
 *		Wake up a single context currently waiting on the
 *		specified event variable via WAIT_EVENT().
 *
 *	BROADCAST_EVENT(&event)
 *
 *		Wake up all contexts currently waiting on the
 *		specified event variable via WAIT_EVENT().
 *
 *	DEINIT_EVENT(&event)
 *
 *		This macro deallocates any data structures that may be
 *		associated with the event variable.  On some systems
 *		this may be a no-op.
 */
typedef event_t EVENT_T;
#define INIT_EVENT(_event) \
	EVENT_INIT(_event)
#define WAIT_EVENT(_event) \
	EVENT_WAIT(_event, PRIMED)
#define SIGNAL_EVENT(_event) \
	EVENT_SIGNAL(_event, 0)
#define BROADCAST_EVENT(_event) \
	EVENT_BROADCAST(_event, 0)
#define DEINIT_EVENT(_event)


/*
 * Interruptable Event Variable Primitives
 * ---------------------------------------
 *
 * The following interfaces are provided for event variables
 * that can potentially be interrupted by POSIX signals:
 *
 *	INIT_INTR_EVENT(&event)
 *
 *		This macro initializes the specified event
 *		variable.  Note that this macro can sleep.
 *
 *	WAIT_INTR_EVENT(&event, is_intr, &was_intr)
 *
 *		Sleep waiting for a certain event to be posted
 *		or a signal to interrupt the wait.  The is_intr
 *		parameter is a boolean specifying whether the
 *		wait is interruptible by POSIX signals.  The
 *		was_intr flag specifies whether the wait was
 *		interrupted by POSIX signals (B_TRUE is returned)
 *		or woken up by either SIGNAL_INTR_EVENT() or
 *		BROADCAST_INTR_EVENT() (B_FALSE is returned).
 *
 *	SIGNAL_INTR_EVENT(&event)
 *
 *		Wake up a single context currently waiting on the
 *		specified event variable via WAIT_INTR_EVENT().
 *
 *	BROADCAST_INTR_EVENT(&event)
 *
 *		Wake up all contexts currently waiting on the
 *		specified event variable via WAIT_INTR_EVENT().
 *
 *	DEINIT_INTR_EVENT(&event)
 *
 *		This macro deallocates any data structures that may be
 *		associated with the event variable.  On some systems
 *		this may be a no-op.
 */
typedef event_t INTR_EVENT_T;
#define INIT_INTR_EVENT(_event) \
	EVENT_INIT(_event)
#define WAIT_INTR_EVENT(_event, _is_intr, _was_intr) \
	do { \
		if (_is_intr) { \
			*(_was_intr) = (EVENT_WAIT_SIG(_event, PRIMED) \
						? B_FALSE : B_TRUE); \
		} \
		else { \
			EVENT_WAIT(_event, PRIMED); \
			*(_was_intr) = B_FALSE; \
		} \
	} while(0)
#define SIGNAL_INTR_EVENT(_event) \
	EVENT_SIGNAL(_event, 0)
#define BROADCAST_INTR_EVENT(_event) \
	EVENT_BROADCAST(_event, 0)
#define DEINIT_INTR_EVENT(_event)


/*
 * Swapout Prevention
 * ------------------
 *
 * Some systems (notably SVR4.2/UnixWare) will swap out a process which
 * is holding critical resources like sleep locks.  This can obviously
 * really kill performance, as you wait for a process to be swapped back
 * in.  These macros are meant to control this sort of swapping behavior.
 * If your OS is smart enough not to swap a process in these conditions,
 * you don't need to bother.
 */
#ifdef NSC_UW
#define	NSC_SWAPLOCK()		(++(u.u_lwpp->l_keepcnt))
#define	NSC_SWAPUNLOCK()	(--(u.u_lwpp->l_keepcnt))
#else  /* !NSC_UW */
#define	NSC_SWAPLOCK()		ATOMIC_INT_INCR(&(u.u_procp->p_lwpkeepcnt))
#define	NSC_SWAPUNLOCK()	ATOMIC_INT_DECR(&(u.u_procp->p_lwpkeepcnt))
#endif /* !NSC_UW */

#endif /* !_NSC_GENERIC_NSC_SYNCH_H */
